SystemOrganization addCategory: #'Ma client server-Client API'!
SystemOrganization addCategory: #'Ma client server-Exceptions'!
SystemOrganization addCategory: #'Ma client server-Info'!
SystemOrganization addCategory: #'Ma client server-Server API'!

MaObject subclass: #MaClientConnection
	instanceVariableNames: 'serverSocket lifecycle extractProcess shouldContinue'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma client server-Server API'!

!MaClientConnection commentStamp: 'cmm 11/13/2008 15:34' prior: 0!
A MaClientConnection is created in the server image when a MaClientSocket connects to a MaServerSocket.  At that point, the client can send request ByteArrays which will be extracted by my extractProcess.  The server allows up to 50 simultaneous connections, after which the oldest is disconnected (but not while they're in a request-lifecycle, of course, since they have to be able to send the response).

Instance Variables
	extractProcess:		<Process> - loops while the client #socket is connected, listening for request ByteArrays, extracting them and processing them when they become available.
	lifecycle:		<Object>
	serverSocket:		<Object>
	shouldContinue:		Set false to allow the extract process and associated socket to clean up normally.
!

!MaClientConnection class methodsFor: 'create' stamp: 'cmm 6/8/2008 10:46'!
serverSocket: aMaServerSocket lifecycle: aMaServerRequestLifecycle 
	^ self new
		serverSocket: aMaServerSocket ;
		lifecycle: aMaServerRequestLifecycle ;
		yourself! !

!MaClientConnection methodsFor: 'private' stamp: 'cmm 6/8/2008 10:51'!
ensureByteArray: aByteArray canHold: anInteger 
	| newAnswer |
	^ aByteArray size < anInteger 
		ifTrue: 
			[ newAnswer := ByteArray new: anInteger.
			newAnswer 
				replaceFrom: 1
				to: aByteArray size
				with: aByteArray
				startingAt: 1.
			newAnswer ]
		ifFalse: [ aByteArray ]! !

!MaClientConnection methodsFor: 'initialize-release' stamp: 'cmm 6/8/2008 10:44'!
initialize
	super initialize.
	shouldContinue := true! !

!MaClientConnection methodsFor: 'private' stamp: 'cmm 6/29/2008 16:26'!
isCompressedBitSet: anInteger 
	^ anInteger maIsBitSet: 32! !

!MaClientConnection methodsFor: 'initialize-release' stamp: 'cmm 11/9/2008 20:15'!
lifecycle: aMaServerRequestLifecycle 
	lifecycle := aMaServerRequestLifecycle.
	lifecycle clientConnection: self! !

!MaClientConnection methodsFor: 'actions' stamp: 'cmm 9/21/2008 17:38'!
monitorForByteArrayRequests
	"A socket connection has just been established.  Kickoff a new iterating process for the purpose of listening for requests bytes from a single client (my lifecycles #socket).  receiveOneByteArrayRequestFrom: waits a few seconds on the port, if data comes in it will extract the entire request and return it.  Otherwise it returns nil and the loop can check for shutdown or expiry."
	extractProcess := 
	[ [ self shouldContinue and: [ lifecycle socket isConnected ] ] whileTrue: 
		[ | requestByteArray |
		requestByteArray := [ self receiveOneByteArrayRequestFrom: lifecycle ] 
			on: Error
			do: 
				[ : exc | 
				self server 
					noteException: exc
					whileReceivingRequestFrom: lifecycle.
				self shouldContinue: false.
				nil ].
		requestByteArray ifNotNil: 
			[ | newLifecycle |
			newLifecycle := lifecycle copyForNewRequest
				requestByteArray: requestByteArray ;
				yourself.
			serverSocket queueRequest: newLifecycle ] ].
	lifecycle socket closeAndDestroy: serverSocket timeoutSeconds.
	serverSocket removeConnection: self ] newProcess.
	extractProcess
		name: 'extracting requests from ' , lifecycle remoteAddress asString , ':' , lifecycle remotePort asString ;
		priority: Processor userInterruptPriority ;
		resume! !

!MaClientConnection methodsFor: 'actions' stamp: 'cmm 11/9/2008 23:42'!
monitorForUdpByteArrayRequests
	"Kickoff a new iterating process for the purpose of listening for requests bytes from UDP clients.  receiveOneByteArrayRequestFrom: waits a few seconds on the port, if data comes in it will extract the entire request and return it.  Otherwise it returns nil and the loop can check for shutdown or expiry."
	extractProcess := 
	[ [ self shouldContinue ] whileTrue: 
		[ | requestByteArray |
		requestByteArray := [ self receiveOneUdpRequestFrom: lifecycle ] 
			on: Error
			do: 
				[ : exc | 
				self server 
					noteException: exc
					whileReceivingRequestFrom: lifecycle.
				self shouldContinue: false.
				nil ].
		requestByteArray ifNotNil: 
			[ | newLifecycle |
			newLifecycle := lifecycle copyForNewRequest
				requestByteArray: requestByteArray ;
				yourself.
			serverSocket queueRequest: newLifecycle ] ] ] newProcess.
	extractProcess
		name: 'extracting requests from UDP clients' ;
		priority: Processor userInterruptPriority ;
		resume! !

!MaClientConnection methodsFor: 'private' stamp: 'cmm 9/21/2008 22:00'!
receiveOneByteArrayRequestFrom: aMaRequestLifecycle 
	"Wait up to 3 seconds for something from the client.  If data is received from aMaRequestLifecycle's socket be sure the entire data is received.  Otherwise answer nil."
	| bytesExpected bytesReceived answer firstDataReceived compressed |
	bytesReceived := 0.
	bytesExpected := nil.
	answer := ByteArray new: 80.
	[ bytesExpected isNil or: [ bytesReceived < bytesExpected ] ] whileTrue: 
		[ aMaRequestLifecycle socket 
			waitForDataFor: 3 "seconds so server can respond to shutdown within 5 seconds total."
			ifClosed: 
				[ ConnectionClosed signal: 'Connection closed while waiting for data.' ]
			ifTimedOut: 
				[ "no data, just fall out"
				 ].
		aMaRequestLifecycle socket dataAvailable ifFalse: 
			[ "This logic means, if data starts to be received, each chunk must be received within 3 seconds or else it will timeout."
			bytesReceived = 0 
				ifTrue: [ ^ nil ]
				ifFalse: 
					[ ConnectionTimedOut signal: 'data not available from ' , (serverSocket hostAndPortStringFor: aMaRequestLifecycle socket) ] ].
		firstDataReceived := Time millisecondClockValue.
		bytesReceived := bytesReceived + (aMaRequestLifecycle socket 
				receiveSomeDataInto: answer
				startingAt: bytesReceived + 1).
		bytesExpected ifNil: 
			[ bytesReceived >= serverSocket headerSize ifTrue: 
				[ bytesExpected := answer 
					maUint: 32
					at: 0.
				compressed := self isCompressedBitSet: bytesExpected.
				bytesExpected := bytesExpected maBitClear: 32.
				bytesExpected > MaObjectBuffer unreasonablePhysicalSize ifTrue: [ MaMalformedRequestError signal: 'unreasonable request size' ].
				answer := self 
					ensureByteArray: answer
					canHold: bytesExpected ] ] ].
	bytesReceived = bytesExpected ifFalse: 
		[ NetworkError signal: 'Unexpected client behavior, more bytes received than client stated sending.' ].
	self server noteTimeToExtractRequestBytes: (Time millisecondsSince: firstDataReceived).
	self server noteTimeToLoadBuffers: (Time millisecondsToRun: 
			[ answer := answer 
				copyFrom: serverSocket headerSize + 1
				to: bytesExpected.
			compressed ifTrue: [ answer := answer lzUncompressed ] ]).
	^ answer! !

!MaClientConnection methodsFor: 'private' stamp: 'cmm 11/9/2008 23:40'!
receiveOneUdpRequestFrom: aMaRequestLifecycle 
	"Wait up to 3 seconds for something from any UDP client.  If data is received from aMaRequestLifecycle's socket, answer it, otherwise nil."
	| bytesExpected bytesReceived compressed receiveArray answer skt |
	bytesReceived := 0.
	bytesExpected := nil.
	skt := aMaRequestLifecycle socket.
	skt 
		waitForDataFor: 3
		ifClosed: 
			[ ConnectionClosed signal: 'Connection closed while waiting for data.' ]
		ifTimedOut: 
			[ "no data, just fall out"
			 ].
	^ aMaRequestLifecycle socket dataAvailable ifTrue: 
		[ answer := ByteArray new: MaClientSocket mtuSize.
		receiveArray := skt receiveUDPDataInto: answer.
		bytesReceived := receiveArray first.
		bytesReceived > serverSocket headerSize ifTrue: 
			[ bytesExpected := answer 
				maUint: 32
				at: 0.
			compressed := self isCompressedBitSet: bytesExpected.
			bytesExpected := bytesExpected maBitClear: 32.
			bytesExpected > MaClientSocket mtuSize ifTrue: [ ^ nil ].
			self server noteTimeToLoadBuffers: 
				(Time millisecondsToRun: 
					[ answer := answer 
						copyFrom: serverSocket headerSize + 1
						to: bytesExpected.
					compressed ifTrue: [ answer := answer lzUncompressed ] ]).
			answer ] ]! !

!MaClientConnection methodsFor: 'private' stamp: 'cmm 6/8/2008 10:47'!
server
	^ serverSocket server! !

!MaClientConnection methodsFor: 'initialize-release' stamp: 'cmm 6/8/2008 10:45'!
serverSocket: aMaServerSocket
	serverSocket := aMaServerSocket! !

!MaClientConnection methodsFor: 'accessing' stamp: 'cmm 6/8/2008 10:43'!
shouldContinue
	^ shouldContinue and: [ serverSocket isRunning ]! !

!MaClientConnection methodsFor: 'accessing' stamp: 'cmm 6/8/2008 11:40'!
shouldContinue: aBoolean
	shouldContinue := aBoolean! !

MaObject subclass: #MaClientServerPreferences
	instanceVariableNames: 'timeoutSeconds compressThreshold'
	classVariableNames: 'Debug'
	poolDictionaries: ''
	category: 'Ma client server-Server API'!

!MaClientServerPreferences commentStamp: 'cmm 2/1/2005 15:27' prior: 0!
Control your Ma client server preferences with this class.!

!MaClientServerPreferences class methodsFor: 'accessing' stamp: 'cmm 1/2/2003 14:35'!
debug

	^Debug! !

!MaClientServerPreferences class methodsFor: 'accessing' stamp: 'cmm 1/2/2003 14:53'!
debug: aBoolean

	Debug _ aBoolean! !

!MaClientServerPreferences class methodsFor: 'private' stamp: 'cmm 1/2/2003 14:35'!
initialize

	super initialize.
	Debug _ false! !

!MaClientServerPreferences methodsFor: 'accessing' stamp: 'cmm 7/5/2008 23:02'!
compressThreshold
	"The maximum number of bytes a request will be before the client decides it should be compressed."
	^compressThreshold! !

!MaClientServerPreferences methodsFor: 'accessing' stamp: 'cmm 7/5/2008 22:48'!
compressThreshold: anInteger
	"The maximum number of bytes a request will be before the client decides it should be compressed."
	compressThreshold := anInteger! !

!MaClientServerPreferences methodsFor: 'initialize-release' stamp: 'cmm 7/5/2008 22:48'!
initialize
	super initialize.
	compressThreshold := 200.
	timeoutSeconds := 10! !

!MaClientServerPreferences methodsFor: 'accessing' stamp: 'cmm 7/5/2008 22:48'!
timeoutSeconds
	^timeoutSeconds! !

!MaClientServerPreferences methodsFor: 'accessing' stamp: 'cmm 7/5/2008 22:52'!
timeoutSeconds: anInteger 
	"The number of seconds to wait for responses from network requests before assuming they are lost."
	timeoutSeconds := anInteger! !

MaObject subclass: #MaClientServerRequest
	instanceVariableNames: 'lifecycle'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma client server-Client API'!

!MaClientServerRequest commentStamp: 'cmm 9/20/2008 16:45' prior: 0!
An abstract superclass for all requests in the "Ma client server" framework.  This superclass knows the #clientConnection, a MaClientConnection, which may be useful to the application server program.!

!MaClientServerRequest methodsFor: 'accessing' stamp: 'cmm 9/21/2008 15:45'!
clientConnection
	^ self lifecycle ifNotNilDo: [ : lc | lc clientConnection ]! !

!MaClientServerRequest methodsFor: 'accessing' stamp: 'cmm 9/20/2008 17:08'!
lifecycle
	^lifecycle! !

!MaClientServerRequest methodsFor: 'accessing' stamp: 'cmm 9/20/2008 17:08'!
lifecycle: aMaServerRequestLifecycle
	lifecycle := aMaServerRequestLifecycle! !

!MaClientServerRequest methodsFor: 'testing' stamp: 'cmm 10/28/2008 14:34'!
wantsResponse
	"Answer whether this request expects a response from the server.  If false, submitting this request will be sent over the socket and immediately answer nil.  The server will make the same test to detemine whether to bother sending a response.
	If true, submitting this request will block and wait for a response from the server."
	"Default is true, subclasses may override."
	^ true! !

MaObject subclass: #MaClientSocket
	instanceVariableNames: 'location socket byteArray stat preferences networkProtocol'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma client server-Client API'!

!MaClientSocket commentStamp: 'cmm 2/1/2005 15:29' prior: 0!
An easy-to-use class to send ByteArray or String requests to a socket listening on the network.  The other socket is listening using MaServerSocket.!

!MaClientSocket class methodsFor: 'initializing' stamp: 'cmm 5/2/2003 15:08'!
initialize

	Smalltalk addToStartUpList: self! !

!MaClientSocket class methodsFor: 'creation' stamp: 'cmm 7/5/2008 14:21'!
location: aMaServerLocation 
	Socket initializeNetwork.
	^ self new
		location: aMaServerLocation ;
		yourself! !

!MaClientSocket class methodsFor: 'private' stamp: 'cmm 11/9/2008 20:03'!
mtuSize
	^ 1500! !

!MaClientSocket class methodsFor: 'initializing' stamp: 'cmm 5/2/2003 15:11'!
startUp: aBoolean

	super startUp: aBoolean.
	Socket initializeNetwork! !

!MaClientSocket methodsFor: 'accessing' stamp: 'cmm 7/5/2008 22:51'!
compressThreshold
	"The maximum number of bytes a request will be before the client decides it should be compressed."
	^ preferences compressThreshold ! !

!MaClientSocket methodsFor: 'accessing' stamp: 'cmm 7/5/2008 22:51'!
compressThreshold: anInteger 
	"The maximum number of bytes a request will be before the client decides it should be compressed."
	preferences compressThreshold: anInteger! !

!MaClientSocket methodsFor: 'private' stamp: 'cmm 11/11/2008 14:46'!
connect
	| startTryingToConnectSocket |
	socket 
		ifNotNil: 
			[ MagmaSoftwareError signal: 'client trying to make multiple connects?' ]
		ifNil: 
			[ self isUdp 
				ifTrue: 
					[ "Do NOT set the port now or it will not work.  It is specified on every send call."
					socket := Socket newUDP ]
				ifFalse: 
					[ socket := Socket newTCP.
					startTryingToConnectSocket := Time millisecondClockValue.
					socket 
						connectTo: self hostAddress
						port: self port
						waitForConnectionFor: self timeoutSeconds.
					stat timeToConnectSocketHistory addValue: (Time millisecondsSince: startTryingToConnectSocket) ] ]! !

!MaClientSocket methodsFor: 'private' stamp: 'cmm 11/24/2008 12:29'!
disconnect
	stat release.
	socket ifNotNil: 
		[ socket closeAndDestroy: 5.
		socket := nil ]! !

!MaClientSocket methodsFor: 'private' stamp: 'cmm 1/23/2009 13:28'!
ensureByteArrayCanHold: anInteger 
	anInteger > 10000000 ifTrue: 
		[ MaUserError signal: 'Client/server transmissions are arbitrarily limited to a maximum of 10-million bytes.' ].
	anInteger > byteArray size ifTrue: 
		[ | newByteArray |
		newByteArray := ByteArray new: anInteger.
		newByteArray 
			replaceFrom: 1
			to: byteArray size
			with: byteArray
			startingAt: 1.
		byteArray := newByteArray ]! !

!MaClientSocket methodsFor: 'private' stamp: 'cmm 6/2/2008 15:53'!
ensureConnected
	"Ensure the primitive Socket is connected."
	socket 
		ifNil: [ self connect ]
		ifNotNil: 
			[ socket isConnected ifFalse: 
				[ self
					disconnect ;
					connect ] ]! !

!MaClientSocket methodsFor: 'private' stamp: 'cmm 9/20/2002 14:28'!
headerSize

	^4! !

!MaClientSocket methodsFor: 'accessing' stamp: 'cmm 7/5/2008 13:26'!
hostAddress
	^ location hostAddress! !

!MaClientSocket methodsFor: 'private' stamp: 'cmm 7/5/2008 15:39'!
hostAndPortString
	^ String streamContents: [ : stream | location maPrintAbbreviatedOn: stream ]! !

!MaClientSocket methodsFor: 'initializing' stamp: 'cmm 7/5/2008 22:50'!
initialize
	super initialize.
	byteArray := ByteArray new: 1000.
	stat := MaClientStat new.
	preferences := MaClientServerPreferences new! !

!MaClientSocket methodsFor: 'private' stamp: 'cmm 6/29/2008 10:50'!
isCompressedBitSet: anInteger 
	^ anInteger maIsBitSet: 32! !

!MaClientSocket methodsFor: 'testing' stamp: 'cmm 9/21/2008 17:53'!
isConnected
	^ socket notNil and: [ socket isConnected ]! !

!MaClientSocket methodsFor: 'testing' stamp: 'cmm 11/11/2008 14:46'!
isUdp
	"Answer true if using a UDP interface, false if TCP/IP."
	^ networkProtocol == #udp! !

!MaClientSocket methodsFor: 'private' stamp: 'cmm 11/9/2008 10:13'!
loadByteArrayWith: aByteArray startingAt: indexPosition count: anInteger 
	"Load my byteArray with aByteArray, compressing if it appropriate."
	| sendSize byteArrayToSend compressed header |
	byteArrayToSend := anInteger > self compressThreshold 
		ifTrue: 
			[ compressed := aByteArray size > anInteger 
				ifTrue: [ (aByteArray 
					copyFrom: 1
					to: anInteger) lzCompressed ]
				ifFalse: [ aByteArray lzCompressed ].
			compressed size < anInteger 
				ifTrue: 
					[ sendSize := self headerSize + compressed size.
					compressed ]
				ifFalse: 
					[ sendSize := self headerSize + anInteger.
					compressed := nil.
					aByteArray ] ]
		ifFalse: 
			[ sendSize := self headerSize + anInteger.
			aByteArray ].
	self ensureByteArrayCanHold: sendSize.
	byteArray 
		replaceFrom: self headerSize + 1
		to: sendSize
		with: byteArrayToSend
		startingAt: indexPosition.
	"If compressed, indicate so in the high-bit of the header."
	header := self 
		setCompressedBit: compressed notNil
		for: sendSize.
	byteArray 
		maUint: 32
		at: 0
		put: header.
	^ sendSize! !

!MaClientSocket methodsFor: 'accessing' stamp: 'cmm 7/5/2008 15:19'!
location
	^ location! !

!MaClientSocket methodsFor: 'initializing' stamp: 'cmm 7/5/2008 14:21'!
location: aMaServerLocation
	location := aMaServerLocation! !

!MaClientSocket methodsFor: 'printing' stamp: 'cmm 11/19/2008 14:33'!
maPrintAbbreviatedOn: aStream 
	super maPrintAbbreviatedOn: aStream.
	location maPrintAbbreviatedOn: aStream! !

!MaClientSocket methodsFor: 'accessing' stamp: 'cmm 11/11/2008 14:37'!
networkProtocol
	^ networkProtocol! !

!MaClientSocket methodsFor: 'accessing' stamp: 'cmm 11/11/2008 14:36'!
networkProtocol: aSymbol
	networkProtocol := aSymbol! !

!MaClientSocket methodsFor: 'accessing' stamp: 'cmm 9/21/2008 16:16'!
port
	^ location ifNotNil: [ location port ]! !

!MaClientSocket methodsFor: 'private' stamp: 'cmm 9/21/2008 22:02'!
receiveInto: aByteArray 
	"First, we receive into *my* byteArray, then I paste it into aByteArray."
	| bytesExpected bytesReceived answer startedWaitingForResponse startedReceivingResponse timeToReceiveResponse compressed |
	bytesReceived := 0.
	bytesExpected := nil.
	"We track two separate timing stats from this method, first the #timeWaitingForResponsesHistory is the time to receive the *initial bytes* of the response.  Second, the #timeReceivingResponsesHistory measures the time to actually *download* the response bytes, calculated by subtracting the value last recorded in the timeWaitingForResponsesHistory."
	startedReceivingResponse := Time millisecondClockValue.
	[ bytesExpected isNil or: [ bytesReceived < bytesExpected ] ] whileTrue: 
		[ startedWaitingForResponse ifNil: [ startedWaitingForResponse := Time millisecondClockValue ].
		socket waitForDataFor: self timeoutSeconds.
		startedWaitingForResponse isInteger ifTrue: 
			[ stat timeWaitingForResponsesHistory addValue: (Time millisecondsSince: startedWaitingForResponse).
			startedWaitingForResponse := 'initial wait captured.'	"Deactivate it now by setting it to this string." ].
		
		[ bytesReceived := bytesReceived + (socket 
				receiveSomeDataInto: byteArray
				startingAt: bytesReceived + 1) ] 
			on: Error
			do: 
				[ : err | 
				err messageText = 'a primitive has failed' ifTrue: 
					[ err resignalAs: (NetworkError new
							messageText: 'Network failure while receiving response.' ;
							yourself) ] ].
		bytesExpected ifNil: 
			[ bytesReceived >= self headerSize ifTrue: 
				[ bytesExpected := byteArray 
					maUint: 32
					at: 0.
				compressed := self isCompressedBitSet: bytesExpected.
				bytesExpected := bytesExpected maBitClear: 32.
				self ensureByteArrayCanHold: bytesExpected ] ] ].
	"I think it's safe to assume here that the stats will be active."
	timeToReceiveResponse := (Time millisecondsSince: startedReceivingResponse) - stat timeWaitingForResponsesHistory last.
	stat timeReceivingResponsesHistory addValue: timeToReceiveResponse.
	"We should now have the full buffer in my byteArray, paste it into the answer ByteArray."
	stat timeToLoadBuffersHistory addValue: (Time millisecondsToRun: 
			[ answer := bytesReceived - self headerSize > aByteArray size 
				ifTrue: [ ByteArray new: bytesReceived - self headerSize ]
				ifFalse: [ aByteArray ].
			answer 
				replaceFrom: 1
				to: bytesReceived - self headerSize
				with: byteArray
				startingAt: self headerSize + 1 ]).
	^ compressed 
		ifTrue: [ answer lzUncompressed ]
		ifFalse: [ answer ]! !

!MaClientSocket methodsFor: 'private' stamp: 'cmm 11/11/2008 14:46'!
send: anInteger 
	| startedSending |
	(anInteger > MaClientSocket mtuSize and: [ self isUdp ]) ifTrue: 
		[ MaClientServerUserError signal: '#udp transmissions should be less than ', MaClientSocket mtuSize, ' bytes' ].
	self ensureConnected.
	startedSending := Time millisecondClockValue.
	self isUdp 
		ifTrue: 
			[ socket 
				sendUDPData: byteArray
				toHost: self hostAddress
				port: self port ]
		ifFalse: 
			[ socket 
				sendData: byteArray
				count: anInteger.
			socket waitForSendDoneFor: self timeoutSeconds.
			socket sendDone ifFalse: 
				[ NetworkError signal: 'Connection to ' , self hostAndPortString , ' lost.' ] ].
	"Only include successful sends to measure real network performance, not timeouts perhaps due to user-error."
	stat timeToSendRequestsHistory addValue: (Time millisecondsSince: startedSending)! !

!MaClientSocket methodsFor: 'send' stamp: 'cmm 11/11/2008 14:46'!
sendData: requestByteArray 
	^ self 
		sendData: requestByteArray
		startingAt: 1
		count: requestByteArray size
		waitForReplyIn: (self isUdp ifTrue: [ ByteArray new: 500 ])! !

!MaClientSocket methodsFor: 'send' stamp: 'cmm 11/13/2008 13:42'!
sendData: requestByteArray startingAt: startIndex count: numberOfBytes waitForReplyIn: answerByteArray 
	"Perform a synchronous TCP call, sending count bytes of requestByteArray from position startIndex to my peer socket and then wait up to my timeout period for a reply, which I'll put in answerByteArray.  If I run out of room in answerByteArray, I'll create you a new one and return it, otherwise you'll get your own answerByteArray back.  If communications are disrupted, I'll signal a NetworkError."
	| tries bytesToSend |
	stat timeToLoadBuffersHistory addValue: (Time millisecondsToRun: 
			[ bytesToSend := self 
				loadByteArrayWith: requestByteArray
				startingAt: startIndex
				count: numberOfBytes ]).
	"Do a retry logic to help with the case where the server has dropped only momentarily.  Sockets will be invalid but, unfortunately, they don't always correctly report they are invalid, as if they don't 'realize' it yet."
	tries := 0.
	[ self send: bytesToSend ] 
		on: Error
		do: 
			[ : sendErr | 
			"Handle a specific bug in Squeak, arrggghh!!"
			(sendErr messageText = 'a primitive has failed' or: [ sendErr isKindOf: NetworkError ]) 
				ifTrue: 
					[ tries < 3 
						ifTrue: 
							[ tries := tries + 1.
							sendErr retry ]
						ifFalse: [ sendErr pass ] ]
				ifFalse: [ sendErr pass ] ].
	"Retry logic is not appropriate on the receive side."
	^ answerByteArray ifNotNil: [ self isUdp ifFalse: [ self receiveInto: answerByteArray ] ]! !

!MaClientSocket methodsFor: 'private' stamp: 'cmm 6/29/2008 11:01'!
setCompressedBit: aBoolean for: headerInteger 
	^ aBoolean 
		ifTrue: 
			[ "Below is the correct implementation, but we would need a 2GB ByteArray for the simpler faster implementation below not to work.  2GB can't happen."
			"ba maUint: 32 at: 0 put: ((ba maUint: 32 at: 0) maBitSet: 32)."
			headerInteger maBitSet: 32 ]
		ifFalse: [ headerInteger maBitClear: 32 ]! !

!MaClientSocket methodsFor: 'private' stamp: 'cmm 5/6/2008 23:17'!
stat
	^stat! !

!MaClientSocket methodsFor: 'private' stamp: 'cmm 5/6/2008 23:18'!
stat: aMaClientStat
	stat := aMaClientStat! !

!MaClientSocket methodsFor: 'accessing' stamp: 'cmm 7/5/2008 22:42'!
timeoutSeconds
	"The number of seconds to wait for responses from network requests before assuming they are lost."
	^ preferences timeoutSeconds ! !

!MaClientSocket methodsFor: 'accessing' stamp: 'cmm 7/5/2008 22:50'!
timeoutSeconds: anInteger 
	"The number of seconds to wait for responses from network requests before assuming they are lost."
	preferences timeoutSeconds: anInteger! !

MaObject subclass: #MaServer
	instanceVariableNames: 'console'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma client server-Server API'!

!MaServer commentStamp: 'cmm 2/1/2005 15:35' prior: 0!
Abstract superclass for the "server engine".  I know a console which I send output to.!

MaServer subclass: #MaRequestServer
	instanceVariableNames: 'processor'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma client server-Server API'!

!MaRequestServer commentStamp: 'cmm 2/20/2005 17:15' prior: 0!
Abstract class whose subclass instances handle the server portion of the client-server activity.  The processor should be an object that responds to #value:.  The argument is the actual request object and it should answer the result object.  Requests and responses can be arbitrary objects, not just byte-objects like Strings and ByteArrays, although you must set up your #protocol.!

MaRequestServer subclass: #MaLocalRequestServer
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma client server-Server API'!

!MaLocalRequestServer commentStamp: 'cmm 5/8/2008 21:21' prior: 0!
Use this server as a stub when you are running your object-request server locally within the same image.!

!MaLocalRequestServer class methodsFor: 'creation' stamp: 'cmm 9/19/2004 22:38'!
processor: oneArgBlockOrProcessor

	^ self new 
		processor: oneArgBlockOrProcessor ;
		yourself! !

!MaLocalRequestServer methodsFor: 'as yet unclassified' stamp: 'cmm 5/2/2005 22:25'!
isRunning

	^ true! !

MaRequestServer subclass: #MaNetworkRequestServer
	instanceVariableNames: 'socket serializer nextId protocolByteArray'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma client server-Server API'!

!MaNetworkRequestServer commentStamp: 'cmm 2/1/2005 15:40' prior: 0!
The server to use when you want to allow requests and responses to be any object and run on a remote server.!

!MaNetworkRequestServer class methodsFor: 'creation' stamp: 'cmm 10/8/2002 22:50'!
protocol: collectionOfClasses
	"collectionOfClasses are all request classes and all response classes any client can expect to encounter in the full graph of the requests and responses"

	^self new protocol: collectionOfClasses! !

!MaNetworkRequestServer methodsFor: 'accessing' stamp: 'cmm 7/15/2008 22:08'!
compressThreshold
	^ socket compressThreshold! !

!MaNetworkRequestServer methodsFor: 'accessing' stamp: 'cmm 7/15/2008 22:07'!
compressThreshold: anInteger
	socket compressThreshold: anInteger! !

!MaNetworkRequestServer methodsFor: 'private' stamp: 'cmm 10/11/2002 00:24'!
connectionInfo

	^Array
		with: self getNextId
		with: self protocolByteArray! !

!MaNetworkRequestServer methodsFor: 'private' stamp: 'cmm 10/11/2002 00:25'!
getNextId

	^nextId _ nextId + 1! !

!MaNetworkRequestServer methodsFor: 'private' stamp: 'cmm 11/26/2004 13:15'!
initialize

	super initialize.
	socket _
		MaServerSocket new
			server: self;
			yourself.
	nextId _ 1.
	self initializeSerializer! !

!MaNetworkRequestServer methodsFor: 'private' stamp: 'cmm 6/10/2005 12:21'!
initializeSerializer

	serializer _ MaObjectSerializer new! !

!MaNetworkRequestServer methodsFor: 'testing' stamp: 'cmm 2/20/2005 17:18'!
isRunning

	^ socket isRunning! !

!MaNetworkRequestServer methodsFor: 'testing' stamp: 'cmm 11/8/2008 16:58'!
isUdp
	^ socket isUdp! !

!MaNetworkRequestServer methodsFor: 'private' stamp: 'cmm 5/1/2008 18:16'!
materializeRequest: requestByteArray 
	| timeToMaterialize requestObject |
	timeToMaterialize := Time millisecondsToRun: 
		[ serializer resetOidManager.
		requestObject := serializer materializeGraph: requestByteArray ].
	self noteTimeToMaterializeRequest: timeToMaterialize.
	^ requestObject! !

!MaNetworkRequestServer methodsFor: 'accessing' stamp: 'cmm 11/11/2008 14:13'!
networkProtocol
	"Answer #tcp for TCP/IP interfaces, #udp for UDP."
	^ socket networkProtocol! !

!MaNetworkRequestServer methodsFor: 'accessing' stamp: 'cmm 11/11/2008 16:27'!
networkProtocol: aSymbol 
	"Specify #tcp or #udp."
	^ socket networkProtocol: aSymbol! !

!MaNetworkRequestServer methodsFor: 'accessing' stamp: 'cmm 9/20/2008 16:08'!
port
	^ socket port! !

!MaNetworkRequestServer methodsFor: 'start / stop' stamp: 'cmm 11/10/2008 16:03'!
processOn: portNumber 
	"Listen on portNumber in the background.  For each client request received, evaluate my processBlock.  I queue all requests so that your block will never be evaluating more than one client-request at a time."
	socket server: self.
	socket processOn: portNumber! !

!MaNetworkRequestServer methodsFor: 'start / stop' stamp: 'cmm 11/9/2008 11:08'!
processOn: portNumber using: yourProcessor 
	"Listen on portNumber in the background.  For each client request received, send value: to yourProcessor (so yourProcessor must be a Block or respond to #value:).  I queue all requests and ensure your yourProcessor will never be evaluating more than one client-request at a time.  The argument to the value: message is an instance of MaClientServerRequest, of which you create application-specific subclasses."
	processor := yourProcessor.
	self processOn: portNumber! !

!MaNetworkRequestServer methodsFor: 'private' stamp: 'cmm 11/10/2008 15:37'!
processRequestByteArray: aMaServerRequestLifecycle 
	"Materialize the object serialized in requestByteArray, and serialize the response I get from telling my processor to process it."
	| requestObject responseObject |
	[ requestObject := self materializeRequest: aMaServerRequestLifecycle requestByteArray.
	aMaServerRequestLifecycle requestObject: requestObject ] 
		on: Error
		do: 
			[ : exc | 
			MaClientServerPreferences debug 
				ifTrue: [ exc pass ]
				ifFalse: 
					[ ^ self serializeResponse: (MaMalformedRequestError new
							messageText: exc messageText ;
							yourself) ] ].
	self connectionRequest = requestObject ifFalse: [ requestObject lifecycle: aMaServerRequestLifecycle ].
	responseObject := self processRequest: requestObject.
	^ (requestObject wantsResponse and: [ self isUdp not ]) ifTrue: [ self serializeResponse: responseObject ]! !

!MaNetworkRequestServer methodsFor: 'private' stamp: 'cmm 8/23/2007 16:26'!
protocol: aCollection
	"Establish the class ids of classes in the protocol."
	aCollection do: [ : each | serializer classIdManager addNewClass: each ].
	"Protocol established, now disallow any further new classes!!"
	serializer classIdManager beImmutable ! !

!MaNetworkRequestServer methodsFor: 'private' stamp: 'cmm 5/1/2008 18:24'!
protocolByteArray
	^ protocolByteArray ifNil: [ protocolByteArray := serializer classDefinitionsByteArray ]! !

!MaNetworkRequestServer methodsFor: 'private' stamp: 'cmm 5/1/2008 19:00'!
serializeResponse: anObject 
	| timeToSerializeResponse response |
	timeToSerializeResponse := Time millisecondsToRun: 
		[ serializer resetOidManager.
		response := (serializer serializeGraph: anObject) trimmedByteArray ].
	console noteTimeToSerializeResponse: timeToSerializeResponse.
	^ response! !

!MaNetworkRequestServer methodsFor: 'start / stop' stamp: 'cmm 11/22/2008 10:51'!
shutdown
	socket shutdown.
	"socket shutdown immediately disallows any more requests, but blocks until pending requests get processed.  Those remaining requests which are not finished writing responses can and will continue to write, but the server is no longer needed and able to shutdown."
	super shutdown
! !

!MaNetworkRequestServer methodsFor: 'accessing' stamp: 'cmm 10/8/2002 22:23'!
timeoutSeconds

	^socket timeoutSeconds! !

!MaNetworkRequestServer methodsFor: 'accessing' stamp: 'cmm 10/8/2002 22:23'!
timeoutSeconds: anInteger

	socket timeoutSeconds: anInteger! !

!MaRequestServer methodsFor: 'private' stamp: 'cmm 5/1/2008 20:55'!
connectionRequest

	^'connect'! !

!MaRequestServer methodsFor: 'initializing' stamp: 'cmm 1/5/2003 23:12'!
initialize

	super initialize.
	console _ MaRequestServerConsole new! !

!MaRequestServer methodsFor: 'testing' stamp: 'cmm 2/20/2005 17:18'!
isRunning

	self subclassResponsibility! !

!MaRequestServer methodsFor: 'private' stamp: 'cmm 9/29/2008 22:38'!
processRequest: aMaClientServerRequest 
	"Here's your request object, do something with it and answer any object, though for remote connections only, every class throughout the entire request and responses graph must be part of your #protocol."
	| answer |
	console noteAboutToProcessRequest: aMaClientServerRequest.
	answer := 
	[ aMaClientServerRequest = self connectionRequest 
		ifTrue: [ self connectionInfo ]
		ifFalse: [ processor value: aMaClientServerRequest ] ] 
		on: MaNotification
		do: 
			[ : noti | 
			console note: noti messageText.
			noti resume ].
	console noteRequestProcessed: aMaClientServerRequest.
	^ answer! !

!MaRequestServer methodsFor: 'accessing' stamp: 'cmm 12/8/2004 13:34'!
processor
	^ processor! !

!MaRequestServer methodsFor: 'private' stamp: 'cmm 9/19/2004 22:10'!
processor: oneArgBlock
	"oneArgBlock takes the request object, answers the response object you want to send back to the client.  For remote, these can only be objects defined in the protocol."

	processor _ oneArgBlock! !

!MaRequestServer methodsFor: 'accessing' stamp: 'cmm 11/21/2008 18:28'!
shutdown
	console noteServerShutdown! !

!MaServer methodsFor: 'console' stamp: 'cmm 10/7/2002 23:45'!
console: aMaServerConsole

	console _ aMaServerConsole! !

!MaServer methodsFor: 'private' stamp: 'cmm 10/8/2002 23:05'!
initialize

	super initialize.
	console _ MaServerConsole new! !

!MaServer methodsFor: 'notifications' stamp: 'cmm 10/10/2002 00:29'!
note: aString
	"The socket is notifying us of a message."

	console note: aString! !

!MaServer methodsFor: 'notifications' stamp: 'cmm 12/17/2008 17:25'!
noteAboutToStart
	console noteAboutToStart! !

!MaServer methodsFor: 'notifications' stamp: 'cmm 5/1/2008 20:47'!
noteClientTimedOut: aMaRequestLifecycle 
	"Network issues, client socket is no longer communicating."
	console noteClientTimedOut: aMaRequestLifecycle! !

!MaServer methodsFor: 'notifications' stamp: 'cmm 5/1/2008 20:53'!
noteException: anException whileReceivingRequestFrom: aMaRequestLifecycle 
	"anException was signaled when trying to read from aSocket.  Maybe they disconnected?"
	console 
		noteException: anException
		whileReceivingRequestFrom: aMaRequestLifecycle! !

!MaServer methodsFor: 'notifications' stamp: 'cmm 12/17/2008 17:26'!
noteJustStarted
	console noteJustStarted! !

!MaServer methodsFor: 'notifications' stamp: 'cmm 9/29/2008 22:47'!
noteLastRequestFinish: clockValue 
	"Note the time the last request finished.  Used to track idle time between requests."
	console noteLastRequestFinish: clockValue! !

!MaServer methodsFor: 'notifications' stamp: 'cmm 10/10/2002 00:30'!
noteNoLongerAcceptingRequests
	"Called when the socket has previously been told to shutdown and can no longer accept requests."

	console noteNoLongerAcceptingRequests! !

!MaServer methodsFor: 'notifications' stamp: 'cmm 10/10/2002 00:30'!
noteNoLongerExtractingRequests
	"Called when the socket has previously been told to shutdown and extracted the last pending sockets request."

	console noteNoLongerExtractingRequests! !

!MaServer methodsFor: 'notifications' stamp: 'cmm 10/10/2002 00:30'!
noteNoLongerProcessingRequests
	"Called when the socket has previously been told to shutdown and has processed the last pending request."

	console noteNoLongerProcessingRequests! !

!MaServer methodsFor: 'notifications' stamp: 'cmm 6/8/2008 11:29'!
noteNumberOfConnections: anInteger 
	"Report the number of known MaClientConnections."
	console noteNumberOfConnections: anInteger! !

!MaServer methodsFor: 'notifications' stamp: 'cmm 5/1/2008 17:54'!
noteRequestDequeued: aMaRequestLifecycle 
	"The response has just been popped off the requestQueue and about to be processed.  Consoles will want to note the time spent waiting in the requestQueue."
	console noteRequestDequeued: aMaRequestLifecycle! !

!MaServer methodsFor: 'notifications' stamp: 'cmm 5/1/2008 17:51'!
noteRequestQueued: aMaRequestLifecycle 
	"The aMaRequestLifecycle's now has a requestByteArray that has been extracted from its socket.  The requestByteArray has now been appended to the end of the requestQueue."
	console noteRequestQueued: aMaRequestLifecycle! !

!MaServer methodsFor: 'notifications' stamp: 'cmm 12/23/2008 19:10'!
noteRequestsQueued: aMaRequestLifecycle 
	"The aMaRequestLifecycle's now has a requestByteArray that has been extracted from its socket.  The requestByteArray has now been appended to the end of the requestQueue."
	console noteRequestsQueued: aMaRequestLifecycle! !

!MaServer methodsFor: 'notifications' stamp: 'cmm 5/1/2008 19:31'!
noteResponseSent: aMaRequestLifecycle 
	"Successfully sent the aByteArray to aSocket."
	console noteResponseSent: aMaRequestLifecycle! !

!MaServer methodsFor: 'notifications' stamp: 'cmm 10/10/2002 00:32'!
noteSendResponse: aByteArray timedOutOn: aSocket
	"aSocket must have disconnected or something, we couldn't get the response to them."

	console
		noteSendResponse: aByteArray
		timedOutOn: aSocket! !

!MaServer methodsFor: 'notifications' stamp: 'cmm 10/10/2002 00:31'!
noteSendingResponse: aByteArray to: aSocket
	"a ByteArray is about to be sent to aSocket."

	console
		noteSendingResponse: aByteArray 
		to: aSocket! !

!MaServer methodsFor: 'notifications' stamp: 'cmm 5/1/2008 17:34'!
noteSocketDequeued: aMaRequestLifecycle 
	"aMaRequestLifecycle's socket has been accepted by the server, and a process forked to extracted its byteArray request."
	console noteSocketDequeued: aMaRequestLifecycle! !

!MaServer methodsFor: 'notifications' stamp: 'cmm 10/10/2002 00:32'!
noteSocketQueued: clientSocket
	"clientSocket has been accepted by the server, and a process forked to extracted its byteArray request."

	console noteSocketQueued: clientSocket! !

!MaServer methodsFor: 'notifications' stamp: 'cmm 5/1/2008 16:44'!
noteTimeToExtractRequestBytes: milliseconds 
	"milliseconds is the amount of time required to extract all of the data out of the socket."
	console noteTimeToExtractRequestBytes: milliseconds! !

!MaServer methodsFor: 'notifications' stamp: 'cmm 6/8/2008 21:40'!
noteTimeToLoadBuffers: milliseconds 
	"The number of milliseconds spent compressing and uncompressing ByteArrays for more efficient network transport."
	console noteTimeToLoadBuffers: milliseconds! !

!MaServer methodsFor: 'notifications' stamp: 'cmm 5/1/2008 18:14'!
noteTimeToMaterializeRequest: milliseconds 
	"milliseconds is the amount of time required to extract all of the data out of the socket."
	console noteTimeToMaterializeRequest: milliseconds! !

!MaServer methodsFor: 'private' stamp: 'cmm 10/28/2008 14:40'!
processRequestByteArray: aByteArray
	"Subclasses will want to override this."
	^ 'default response'! !

MaObject subclass: #MaServerConsole
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma client server-Server API'!

!MaServerConsole commentStamp: 'cmm 2/1/2005 15:34' prior: 0!
I am used to receive messages from the server.!

MaServerConsole subclass: #MaRequestServerConsole
	instanceVariableNames: 'statistics lastRequestStart lastRequestFinish'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma client server-Server API'!

!MaRequestServerConsole commentStamp: 'cmm 2/1/2005 15:34' prior: 0!
I am used to receive messages from the server.!

!MaRequestServerConsole methodsFor: 'initialize-release' stamp: 'cmm 9/29/2008 16:11'!
initialize
	super initialize.
	self initializeStatistics! !

!MaRequestServerConsole methodsFor: 'initialize-release' stamp: 'cmm 9/29/2008 16:11'!
initializeStatistics
	statistics := MaServerStat new! !

!MaRequestServerConsole methodsFor: 'notifications' stamp: 'cmm 9/29/2008 15:34'!
noteAboutToProcessRequest: yourRequestObject 
	lastRequestStart := Time millisecondClockValue! !

!MaRequestServerConsole methodsFor: 'notifications' stamp: 'cmm 9/29/2008 16:54'!
noteClientTimedOut: aMaRequestLifecycle 
	"Network issues, client socket is no longer communicating"
	self statistics recordClientTimeout! !

!MaRequestServerConsole methodsFor: 'notifications' stamp: 'cmm 9/29/2008 16:15'!
noteException: anException whileReceivingRequestFrom: aMaRequestLifecycle 
	"anException was signaled when trying to read from aSocket."
	self noteClientTimedOut: aMaRequestLifecycle! !

!MaRequestServerConsole methodsFor: 'notifications' stamp: 'cmm 9/29/2008 22:50'!
noteLastRequestFinish: millisecondClockValueOrNil
	"Note the time the last request finished.  Used to track idle time between requests."
	lastRequestFinish := millisecondClockValueOrNil! !

!MaRequestServerConsole methodsFor: 'notifications' stamp: 'cmm 9/29/2008 16:16'!
noteNumberOfConnections: anInteger 
	"Report the number of known MaClientConnections."
	self statistics numberOfClientConnectionsHistory addValue: anInteger! !

!MaRequestServerConsole methodsFor: 'notifications' stamp: 'cmm 12/23/2008 19:17'!
noteRequestDequeued: aMaRequestLifecycle 
	"Indicates aSocket has delivered a ByteArray request that is about to be processed by my server."
	super noteRequestDequeued: aMaRequestLifecycle.
	self statistics isActive ifTrue: 
		[ self statistics
			recordTimeWaitingInRequestQueue: aMaRequestLifecycle timeWaitingInRequestQueue.
		"Only record idleTime if there was a wait since the last request."
		lastRequestFinish ifNotNil: 
			[ self statistics recordIdleTime: 
				(Time 
					milliseconds: aMaRequestLifecycle requestQueueTime
					since: lastRequestFinish).
			self noteLastRequestFinish: nil ] ]! !

!MaRequestServerConsole methodsFor: 'notifications' stamp: 'cmm 9/29/2008 16:16'!
noteRequestProcessed: requestObject 
	| time |
	time := Time millisecondsSince: lastRequestStart.
	(self statistics requestProcessingTimeHistoryFor: requestObject) addValue: time.
	self statistics requestProcessingTimeHistory addValue: time! !

!MaRequestServerConsole methodsFor: 'notifications' stamp: 'cmm 12/23/2008 19:23'!
noteRequestQueued: aMaRequestLifecycle 
	"The aMaRequestLifecycle's now has a requestByteArray that has been extracted from its socket.  The requestByteArray has now been appended to the end of the requestQueue."
	self statistics isActive ifTrue: 
		[ self statistics recordRequestSize: aMaRequestLifecycle requestByteArray size ]! !

!MaRequestServerConsole methodsFor: 'notifications' stamp: 'cmm 12/23/2008 19:11'!
noteRequestsQueued: anInteger 
	self statistics isActive ifTrue: [ self statistics requestsQueuedHistory addValue: anInteger ]! !

!MaRequestServerConsole methodsFor: 'notifications' stamp: 'cmm 9/29/2008 16:16'!
noteResponseSent: aMaRequestLifecycle 
	"Successfully sent the aMaRequestLifecycle's responseByteArray to its socket."
	self statistics
		recordResponseSize: aMaRequestLifecycle responseByteArray size ;
		recordTimeToSendResponse: aMaRequestLifecycle timeToSendResponse ;
		recordSuccessfulLifecycle! !

!MaRequestServerConsole methodsFor: 'notifications' stamp: 'cmm 11/21/2008 18:16'!
noteServerShutdown
	super noteServerShutdown.
	self statistics
		release ;
		stopTime: DateAndTime now! !

!MaRequestServerConsole methodsFor: 'notifications' stamp: 'cmm 9/29/2008 16:21'!
noteSocketDequeued: aMaRequestLifecycle 
	self statistics isActive ifTrue: 
		[ self statistics
			decrementSocketsQueued ;
			recordTimeWaitingInSocketQueue: aMaRequestLifecycle timeWaitingInSocketQueue ]! !

!MaRequestServerConsole methodsFor: 'notifications' stamp: 'cmm 9/29/2008 16:22'!
noteSocketQueued: aSocket 
	self statistics isActive ifTrue: [ self statistics incrementSocketsQueued ]! !

!MaRequestServerConsole methodsFor: 'notifications' stamp: 'cmm 9/29/2008 16:22'!
noteTimeToExtractRequestBytes: milliseconds 
	"Time to extract the request bytes out of the primitive TCP/IP socket."
	self statistics recordTimeToExtractRequestBytes: milliseconds! !

!MaRequestServerConsole methodsFor: 'notifications' stamp: 'cmm 9/29/2008 16:22'!
noteTimeToLoadBuffers: milliseconds 
	"The number of milliseconds spent compressing and uncompressing ByteArrays for more efficient network transport."
	self statistics timeToLoadBuffersHistory addValue: milliseconds! !

!MaRequestServerConsole methodsFor: 'notifications' stamp: 'cmm 9/29/2008 16:22'!
noteTimeToMaterializeRequest: milliseconds 
	"The time to materialize a serialized request object."
	self statistics timeToMaterializeHistory addValue: milliseconds! !

!MaRequestServerConsole methodsFor: 'notifications' stamp: 'cmm 9/29/2008 16:22'!
noteTimeToSerializeResponse: milliseconds 
	"The time to materialize a serialized request object."
	self statistics timeToSerializeHistory addValue: milliseconds! !

!MaRequestServerConsole methodsFor: 'access' stamp: 'cmm 9/29/2008 16:23'!
statistics
	^ statistics ! !

!MaServerConsole methodsFor: 'logging' stamp: 'cmm 10/25/2008 14:54'!
log: aString
	"Write aString to a log file?"
self halt: aString! !

!MaServerConsole methodsFor: 'notifications' stamp: 'cmm 10/7/2002 23:51'!
note: aString
	"The socket is notifying us of a message."! !

!MaServerConsole methodsFor: 'notifications' stamp: 'cmm 9/29/2008 16:09'!
noteAboutToProcessRequest: yourRequestObject 
	"optional subclass responsibility"! !

!MaServerConsole methodsFor: 'notifications' stamp: 'cmm 12/17/2008 17:26'!
noteAboutToStart
	"For application hook."! !

!MaServerConsole methodsFor: 'notifications' stamp: 'cmm 5/1/2008 20:44'!
noteClientTimedOut: aMaRequestLifecycle 
	"Network issues , client socket is no longer communicating"! !

!MaServerConsole methodsFor: 'notifications' stamp: 'cmm 6/5/2008 21:27'!
noteException: anException whileReceivingRequestFrom: aMaRequestLifecycle
	"anException was signaled when trying to read from aSocket."! !

!MaServerConsole methodsFor: 'notifications' stamp: 'cmm 12/17/2008 17:26'!
noteJustStarted
	"For application hook."! !

!MaServerConsole methodsFor: 'notifications' stamp: 'cmm 9/29/2008 22:48'!
noteLastRequestFinish: millisecondClockValue 
	"Note the time the last request finished.  Used to track idle time between requests."! !

!MaServerConsole methodsFor: 'notifications' stamp: 'cmm 10/7/2002 23:51'!
noteNoLongerAcceptingRequests
	"Called when the socket has previously been told to shutdown and can no longer accept requests."! !

!MaServerConsole methodsFor: 'notifications' stamp: 'cmm 10/7/2002 23:51'!
noteNoLongerExtractingRequests
	"Called when the socket has previously been told to shutdown and extracted the last pending sockets request."! !

!MaServerConsole methodsFor: 'notifications' stamp: 'cmm 10/7/2002 23:51'!
noteNoLongerProcessingRequests
	"Called when the socket has previously been told to shutdown and has processed the last pending request."! !

!MaServerConsole methodsFor: 'notifications' stamp: 'cmm 6/8/2008 11:30'!
noteNumberOfConnections: anInteger 
	"Report the number of known MaClientConnections."! !

!MaServerConsole methodsFor: 'notifications' stamp: 'cmm 5/1/2008 17:54'!
noteRequestDequeued: aMaRequestLifecycle 
	! !

!MaServerConsole methodsFor: 'notifications' stamp: 'cmm 9/29/2008 16:16'!
noteRequestProcessed: yourRequestObject
	"optional subclass responsibility"! !

!MaServerConsole methodsFor: 'notifications' stamp: 'cmm 5/1/2008 17:51'!
noteRequestQueued: aMaRequestLifecycle
! !

!MaServerConsole methodsFor: 'notifications' stamp: 'cmm 12/23/2008 19:10'!
noteRequestsQueued: aMaRequestLifecycle
! !

!MaServerConsole methodsFor: 'notifications' stamp: 'cmm 5/1/2008 19:32'!
noteResponseSent: aMaRequestLifecycle
	"Successfully sent the aMaRequestLifecycle's responseByteArray to its socket."! !

!MaServerConsole methodsFor: 'notifications' stamp: 'cmm 10/7/2002 23:51'!
noteSendResponse: aByteArray timedOutOn: aSocket
	"aSocket must have disconnected or something, we couldn't get the response to them."! !

!MaServerConsole methodsFor: 'notifications' stamp: 'cmm 10/7/2002 23:51'!
noteSendingResponse: aByteArray to: aSocket
	"a ByteArray is about to be sent to aSocket."! !

!MaServerConsole methodsFor: 'notifications' stamp: 'cmm 10/8/2002 23:10'!
noteServerShutdown! !

!MaServerConsole methodsFor: 'notifications' stamp: 'cmm 1/5/2003 18:18'!
noteSocketDequeued: clientSocket
	"clientSocket is about processed by the server."! !

!MaServerConsole methodsFor: 'notifications' stamp: 'cmm 10/7/2002 23:51'!
noteSocketQueued: clientSocket
	"clientSocket has been accepted by the server, and a process forked to extracted its byteArray request."
! !

!MaServerConsole methodsFor: 'notifications' stamp: 'cmm 5/1/2008 16:44'!
noteTimeToExtractRequestBytes: milliseconds
	"Time to extract the request bytes out of the primitive TCP/IP socket."! !

!MaServerConsole methodsFor: 'notifications' stamp: 'cmm 6/8/2008 21:39'!
noteTimeToLoadBuffers: milliseconds
	"The number of milliseconds spent compressing and uncompressing ByteArrays for more efficient network transport."
! !

!MaServerConsole methodsFor: 'notifications' stamp: 'cmm 5/1/2008 18:12'!
noteTimeToMaterializeRequest: milliseconds
	"The time to materialize a serialized request object."! !

!MaServerConsole methodsFor: 'notifications' stamp: 'cmm 5/1/2008 18:55'!
noteTimeToSerializeResponse: milliseconds
	"The time to materialize a serialized request object."! !

MaServerConsole subclass: #MaTranscriptConsole
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma client server-Server API'!

!MaTranscriptConsole commentStamp: 'cmm 2/1/2005 15:41' prior: 0!
A default console that prints messages to the Transcript.!

!MaTranscriptConsole methodsFor: 'output' stamp: 'cmm 10/1/2002 22:49'!
newLine

	Transcript cr! !

!MaTranscriptConsole methodsFor: 'notifications' stamp: 'cmm 10/1/2002 23:33'!
note: aString

	self
		newLine;
		showMessage: aString! !

!MaTranscriptConsole methodsFor: 'notifications' stamp: 'cmm 5/1/2008 20:53'!
noteException: anException onSocket: aMaRequestLifecycle 
	"anException was signaled when trying to read from aSocket.  Maybe they disconnected?"
	self
		newLine ;
		showSocket: aMaRequestLifecycle socket ;
		showSeparator ;
		showMessage: (anException messageText isEmpty 
				ifTrue: [ anException species name ]
				ifFalse: [ anException messageText ])! !

!MaTranscriptConsole methodsFor: 'notifications' stamp: 'cmm 5/1/2008 20:53'!
noteException: anException whileReceivingRequestFrom: aMaRequestLifecycle 
	"anException was signaled when trying to read from aSocket.  Maybe they disconnected?"
	self 
		noteException: anException
		onSocket: aMaRequestLifecycle! !

!MaTranscriptConsole methodsFor: 'notifications' stamp: 'cmm 10/9/2002 23:07'!
noteNoLongerProcessingRequests
	"Called when the socket has previously been told to shutdown and has processed the last pending request."

	self newLine; showMessage: 'I shall process no more requests.'! !

!MaTranscriptConsole methodsFor: 'notifications' stamp: 'cmm 7/16/2008 09:55'!
noteResponseSent: aMaRequestLifecycle 
	"Successfully sent the aMaRequestLifecycle's responseByteArray to its socket."
	self
		newLine ;
		showMessage: 'sent ' , aMaRequestLifecycle responseByteArray size printString , ' bytes to ' , (NetNameResolver stringFromAddress: aMaRequestLifecycle socket remoteAddress)! !

!MaTranscriptConsole methodsFor: 'notifications' stamp: 'cmm 7/16/2008 09:54'!
noteSendResponse: aByteArray timedOutOn: aSocket 
	"aSocket must have disconnected or something, we couldn't get the response to them."
	self
		newLine ;
		showMessage: 'timed out trying to send ' , aByteArray size printString , ' bytes to ' , (NetNameResolver stringFromAddress: aSocket remoteAddress)! !

!MaTranscriptConsole methodsFor: 'notifications' stamp: 'cmm 7/16/2008 09:54'!
noteSendingResponse: aByteArray to: aSocket 
	"a ByteArray is about to be sent to aSocket."
	self
		newLine ;
		showMessage: 'about to send ' , aByteArray size printString , ' bytes to ' , (NetNameResolver stringFromAddress: aSocket remoteAddress)! !

!MaTranscriptConsole methodsFor: 'notifications' stamp: 'cmm 7/16/2008 09:54'!
noteSocketQueued: clientSocket 
	"clientSocket has been accepted by the server, and a process forked to extracted its byteArray request."
	self
		newLine ;
		showMessage: 'queued ' , (NetNameResolver stringFromAddress: clientSocket remoteAddress)! !

!MaTranscriptConsole methodsFor: 'output' stamp: 'cmm 10/1/2002 22:49'!
showMessage: aString

	Transcript show: aString! !

!MaTranscriptConsole methodsFor: 'output' stamp: 'cmm 10/1/2002 22:53'!
showSeparator

	Transcript show: ' - '! !

!MaTranscriptConsole methodsFor: 'output' stamp: 'cmm 7/16/2008 09:53'!
showSocket: aSocket 
	Transcript show: (NetNameResolver stringFromAddress: aSocket remoteAddress)! !

MaObject subclass: #MaServerLink
	instanceVariableNames: 'stat'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma client server-Client API'!

!MaServerLink commentStamp: 'cmm 2/1/2005 15:36' prior: 0!
Abstract superclass representing a clients link to a server.  A client uses this link to submit request objects via #submit:.!

MaServerLink subclass: #MaLocalServerLink
	instanceVariableNames: 'localRequestServer'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma client server-Client API'!

!MaLocalServerLink commentStamp: 'cmm 2/1/2005 15:32' prior: 0!
This is your clients interface to the server when the server is running locally (e.g., with MaLocalRequestServer).!

!MaLocalServerLink class methodsFor: 'as yet unclassified' stamp: 'cmm 3/30/2002 22:50'!
localRequestServer: aMaLocalRequestServer

	^self new localRequestServer: aMaLocalRequestServer
! !

!MaLocalServerLink methodsFor: 'actions' stamp: 'cmm 11/22/2008 12:29'!
disconnect
	super disconnect.
	localRequestServer shutdown! !

!MaLocalServerLink methodsFor: 'testing' stamp: 'cmm 10/8/2002 23:59'!
isLocal

	^true! !

!MaLocalServerLink methodsFor: 'accessing' stamp: 'cmm 11/3/2008 11:47'!
linkInformation
	^ localRequestServer processor pathName! !

!MaLocalServerLink methodsFor: 'private' stamp: 'cmm 7/2/2004 13:17'!
localRequestServer

	^ localRequestServer! !

!MaLocalServerLink methodsFor: 'private' stamp: 'cmm 3/30/2002 22:52'!
localRequestServer: aMaLocalRequestServer

	localRequestServer _ aMaLocalRequestServer
! !

!MaLocalServerLink methodsFor: 'actions' stamp: 'cmm 3/30/2002 22:03'!
submit: aMaServerRequest

	^localRequestServer processRequest: aMaServerRequest
! !

MaServerLink subclass: #MaNetworkServerLink
	instanceVariableNames: 'serializer socket submitGuard'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma client server-Client API'!

!MaNetworkServerLink commentStamp: 'cmm 2/1/2005 15:40' prior: 0!
This is your clients interface to the server when the server is running remotely (e.g., with MaTcpRequestServer).!

!MaNetworkServerLink class methodsFor: 'creation' stamp: 'cmm 7/5/2008 14:29'!
host: hostName port: portInteger 
	^ self location: 
		(MaServerLocation 
			host: hostName
			port: portInteger)! !

!MaNetworkServerLink class methodsFor: 'creation' stamp: 'cmm 7/5/2008 14:18'!
location: aMaServerLocation 
	^ self new
		socket: (MaClientSocket location: aMaServerLocation) ;
		yourself! !

!MaNetworkServerLink methodsFor: 'accessing' stamp: 'cmm 7/5/2008 23:02'!
compressThreshold
	"The maximum number of bytes a request will be before the client decides it should be compressed."
	^ socket compressThreshold! !

!MaNetworkServerLink methodsFor: 'accessing' stamp: 'cmm 7/5/2008 23:03'!
compressThreshold: anInteger 
	"The maximum number of bytes a request will be before the client decides it should be compressed."
	socket compressThreshold: anInteger! !

!MaNetworkServerLink methodsFor: 'connections' stamp: 'cmm 9/20/2008 11:36'!
connect
	"Request and establish protocol from server."
	| idAndProtocol |
	super connect.
	idAndProtocol _ self primSubmit: self connectionRequest.
	serializer classDefinitionsByteArray: (idAndProtocol at: 2).
	serializer classIdManager beImmutable! !

!MaNetworkServerLink methodsFor: 'private' stamp: 'cmm 10/16/2002 14:26'!
connectionRequest

	^'connect'! !

!MaNetworkServerLink methodsFor: 'connections' stamp: 'cmm 5/7/2008 16:49'!
disconnect
	super disconnect.
	socket disconnect! !

!MaNetworkServerLink methodsFor: 'private' stamp: 'cmm 9/23/2008 14:25'!
initialize
	super initialize.
	self initializeSerializer.
	submitGuard := Mutex new! !

!MaNetworkServerLink methodsFor: 'private' stamp: 'cmm 6/10/2005 12:21'!
initializeSerializer

	serializer _ MaObjectSerializer new! !

!MaNetworkServerLink methodsFor: 'connections' stamp: 'cmm 9/21/2008 17:53'!
isConnected
	^ socket notNil and: [ socket isConnected ]! !

!MaNetworkServerLink methodsFor: 'accessing' stamp: 'cmm 11/10/2008 14:07'!
isUdp
	^ socket isUdp! !

!MaNetworkServerLink methodsFor: 'accessing' stamp: 'cmm 7/5/2008 13:30'!
linkInformation
	^ socket hostAndPortString! !

!MaNetworkServerLink methodsFor: 'accessing' stamp: 'cmm 7/5/2008 15:04'!
location
	"Answer the MaServerLocation of the server at the other end of this link."
	^ socket location! !

!MaNetworkServerLink methodsFor: 'printing' stamp: 'cmm 11/12/2008 18:18'!
maPrintAbbreviatedOn: aStream 
	super maPrintAbbreviatedOn: aStream.
	self location maPrintAbbreviatedOn: aStream! !

!MaNetworkServerLink methodsFor: 'accessing' stamp: 'cmm 11/11/2008 14:41'!
networkProtocol
	^ socket networkProtocol! !

!MaNetworkServerLink methodsFor: 'accessing' stamp: 'cmm 11/11/2008 14:37'!
networkProtocol: aSymbol
	socket networkProtocol: aSymbol! !

!MaNetworkServerLink methodsFor: 'private' stamp: 'cmm 11/13/2008 13:58'!
primSubmit: aMaClientServerRequest 
	^ submitGuard critical: 
		[ | answer responseByteArray beganSubmitting graphBuffer timeToSerialize |
		beganSubmitting := Time millisecondClockValue.
		socket stat: stat.
		timeToSerialize := Time millisecondsToRun: 
			[ serializer resetOidManager.
			graphBuffer := serializer serializeGraph: aMaClientServerRequest ].
		stat timeToSerializeHistory addValue: timeToSerialize.
		responseByteArray := socket 
			sendData: graphBuffer byteArray
			startingAt: 1
			count: graphBuffer bufferSize
			waitForReplyIn: ((aMaClientServerRequest wantsResponse and: [ self isUdp not ]) ifTrue: [ serializer byteArray ]).
		responseByteArray ifNotNil: 
			[ serializer resetOidManager.
			stat timeToMaterializeHistory addValue: (Time millisecondsToRun: [ answer := serializer materializeGraph: responseByteArray ]) ].
		(stat timeToSubmitHistoryFor: aMaClientServerRequest class) addValue: (Time millisecondsSince: beganSubmitting).
		answer ]! !

!MaNetworkServerLink methodsFor: 'initialize-release' stamp: 'cmm 11/10/2008 14:10'!
protocol: aCollection 
	"Establish the class ids of classes in the protocol."
	aCollection do: [ : each | serializer classIdManager addNewClass: each ].
	"Protocol established, now disallow any further new classes!!"
	serializer classIdManager beImmutable! !

!MaNetworkServerLink methodsFor: 'connections' stamp: 'cmm 9/20/2008 11:35'!
protocolEstablished
	"Once protocol is established, it is not changed."
	^ serializer classIdManager isMutable not! !

!MaNetworkServerLink methodsFor: 'private' stamp: 'cmm 9/22/2002 21:32'!
socket: aMaClientSocket

	socket _ aMaClientSocket! !

!MaNetworkServerLink methodsFor: 'accessing' stamp: 'cmm 5/7/2008 16:06'!
stat
	^ stat! !

!MaNetworkServerLink methodsFor: 'actions' stamp: 'cmm 12/31/2008 14:08'!
submit: aMaClientServerRequest 
	^ (aMaClientServerRequest wantsResponse or: [ self protocolEstablished ]) 
		ifTrue: 
			[ submitGuard critical: 
				[ "ensure connected"
				self protocolEstablished ifFalse: [ self connect ].
				self primSubmit: aMaClientServerRequest ] ]
		ifFalse: 
			[ 
			[ submitGuard critical: 
				[ self connect.
				self primSubmit: aMaClientServerRequest ] ] newProcess
				name: 'auto-forked unconnected async ' , aMaClientServerRequest printString ;
				resume.
			nil ]! !

!MaNetworkServerLink methodsFor: 'accessing' stamp: 'cmm 10/8/2002 22:23'!
timeoutSeconds

	^socket timeoutSeconds! !

!MaNetworkServerLink methodsFor: 'accessing' stamp: 'cmm 10/8/2002 22:22'!
timeoutSeconds: anInteger

	socket timeoutSeconds: anInteger! !

!MaServerLink methodsFor: 'actions' stamp: 'cmm 6/26/2008 23:05'!
connect
	stat transition! !

!MaServerLink methodsFor: 'actions' stamp: 'cmm 5/7/2008 16:49'!
disconnect
	"Remote links need to clean up resources."
	stat release! !

!MaServerLink methodsFor: 'access' stamp: 'cmm 7/5/2008 14:37'!
host
	MaUserError signal: 'not remotly connected.'! !

!MaServerLink methodsFor: 'initialize-release' stamp: 'cmm 5/7/2008 16:10'!
initialize
	super initialize.
	stat := MaClientStat new! !

!MaServerLink methodsFor: 'testing' stamp: 'cmm 9/20/2008 13:14'!
isConnected
	^ true! !

!MaServerLink methodsFor: 'testing' stamp: 'cmm 10/8/2002 23:59'!
isLocal

	^false! !

!MaServerLink methodsFor: 'override' stamp: 'cmm 7/2/2004 17:16'!
linkInformation

	self subclassResponsibility! !

!MaServerLink methodsFor: 'access' stamp: 'cmm 8/12/2007 20:38'!
port
	MaUserError signal: 'not remotly connected.'! !

!MaServerLink methodsFor: 'testing' stamp: 'cmm 9/20/2008 13:21'!
protocolEstablished
	^ true! !

!MaServerLink methodsFor: 'access' stamp: 'cmm 5/6/2008 23:14'!
stat: aMaClientStat
	stat := aMaClientStat! !

!MaServerLink methodsFor: 'actions' stamp: 'cmm 3/30/2002 22:01'!
submit: aMaServerRequest

	self subclassResponsibility! !

!MaServerLink methodsFor: 'access' stamp: 'cmm 10/16/2002 22:05'!
timeoutSeconds: anInteger
	"for interface compatibility with remote server links."! !

MaObject subclass: #MaServerLocation
	instanceVariableNames: 'host port'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma client server-Client API'!

!MaServerLocation commentStamp: 'cmm 7/5/2008 13:21' prior: 0!
A MaServerLocation simply encapsulates a host and port, allowing the receiver to serve as a 'bookmark' to a server resource.
!

!MaServerLocation class methodsFor: 'create' stamp: 'cmm 7/5/2008 13:20'!
host: aString port: anInteger 
	^ self new 
		setHost: aString
		port: anInteger! !

!MaServerLocation methodsFor: 'testing' stamp: 'cmm 1/11/2009 21:43'!
= aMagmaRemoteLocation 
	aMagmaRemoteLocation maOriginalClass = self species ifFalse: [ ^ false ].
	^ port = aMagmaRemoteLocation port and: 
		[ self hostAddress = aMagmaRemoteLocation hostAddress or: 
			[ self hostAddress maIsLocalHost and: [ aMagmaRemoteLocation hostAddress maIsLocalHost ] ] ]! !

!MaServerLocation methodsFor: 'testing' stamp: 'cmm 1/11/2009 21:43'!
hash
	"Local addresses may be anything that begins with 127."
	| addr |
	addr := self hostAddress maIsLocalHost 
		ifTrue: [ #(127 0 0 1 ) asByteArray ]
		ifFalse: [ self hostAddress ].
	^ addr hash bitXor: port hash! !

!MaServerLocation methodsFor: 'access' stamp: 'cmm 7/5/2008 13:18'!
host
	"The host name of the computer at the remote location."
	^ host! !

!MaServerLocation methodsFor: 'access' stamp: 'cmm 7/5/2008 13:27'!
hostAddress
	"Answer a ByteArray representing the IP address of my remote host."
	^ NetNameResolver 
		addressForName: host
		timeout: 10! !

!MaServerLocation methodsFor: 'printing' stamp: 'cmm 7/5/2008 14:50'!
maPrintAbbreviatedOn: aStream 
	super maPrintAbbreviatedOn: aStream.
	aStream
		maPrint: host ;
		maPrint: $: ;
		maPrint: port! !

!MaServerLocation methodsFor: 'access' stamp: 'cmm 7/5/2008 13:18'!
port
	"Answer the value of port"

	^ port! !

!MaServerLocation methodsFor: 'initialize-release' stamp: 'cmm 11/11/2008 14:32'!
setHost: aString port: anInteger 
	host := aString.
	port := anInteger! !

MaObject subclass: #MaServerRequestLifecycle
	instanceVariableNames: 'socket socketQueueTime socketDequeueTime requestQueueTime requestDequeueTime timeToExtractRequestBytes requestByteArray requestObject beginSendingResponseTime doneSendingResponseTime responseByteArray remoteAddress remotePort clientConnection'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma client server-Server API'!

!MaServerRequestLifecycle class methodsFor: 'create' stamp: 'cmm 11/9/2008 20:15'!
socket: aSocket 
	^ self new
		socket: aSocket ;
		yourself! !

!MaServerRequestLifecycle methodsFor: 'accessing' stamp: 'cmm 5/1/2008 19:19'!
beginSendingResponse
	beginSendingResponseTime := Time millisecondClockValue! !

!MaServerRequestLifecycle methodsFor: 'accessing' stamp: 'cmm 5/1/2008 19:18'!
beginSendingResponseTime
	^beginSendingResponseTime! !

!MaServerRequestLifecycle methodsFor: 'accessing' stamp: 'cmm 5/1/2008 19:18'!
beginSendingResponseTime: anInteger
	beginSendingResponseTime := anInteger! !

!MaServerRequestLifecycle methodsFor: 'accessing' stamp: 'cmm 9/20/2008 17:09'!
clientConnection
	^clientConnection! !

!MaServerRequestLifecycle methodsFor: 'accessing' stamp: 'cmm 9/20/2008 17:09'!
clientConnection: aMaClientConnection
	clientConnection := aMaClientConnection! !

!MaServerRequestLifecycle methodsFor: 'copying' stamp: 'cmm 6/2/2008 13:15'!
copyForNewRequest
	^ self copy postCopyForNewRequest! !

!MaServerRequestLifecycle methodsFor: 'accessing' stamp: 'cmm 5/1/2008 19:26'!
doneSendingResponse
	doneSendingResponseTime := Time millisecondClockValue! !

!MaServerRequestLifecycle methodsFor: 'accessing' stamp: 'cmm 5/1/2008 16:22'!
millisecondsWaitingInSocketQueue
	^ Time 
		milliseconds: socketDequeueTime
		since: socketQueueTime! !

!MaServerRequestLifecycle methodsFor: 'initialize-release' stamp: 'cmm 5/1/2008 16:30'!
noteSocketDequeued
	socketDequeueTime := Time millisecondClockValue! !

!MaServerRequestLifecycle methodsFor: 'copying' stamp: 'cmm 6/2/2008 14:02'!
postCopyForNewRequest
	"The lifecycle stat is instantiated upon the initial socket connection, then cloned for each request to have its own stats."
	requestQueueTime := nil.
	requestDequeueTime := nil.
	timeToExtractRequestBytes := nil.
	requestByteArray := nil.
	beginSendingResponseTime := nil.
	doneSendingResponseTime := nil.
	responseByteArray := nil! !

!MaServerRequestLifecycle methodsFor: 'identification' stamp: 'cmm 5/6/2008 15:39'!
remoteAddress
	^ remoteAddress! !

!MaServerRequestLifecycle methodsFor: 'identification' stamp: 'cmm 5/6/2008 15:39'!
remotePort
	^ remotePort! !

!MaServerRequestLifecycle methodsFor: 'accessing' stamp: 'cmm 5/1/2008 16:37'!
requestByteArray
	^requestByteArray! !

!MaServerRequestLifecycle methodsFor: 'accessing' stamp: 'cmm 7/5/2008 19:43'!
requestByteArray: aByteArray
	requestByteArray := aByteArray! !

!MaServerRequestLifecycle methodsFor: 'initialize-release' stamp: 'cmm 5/1/2008 16:24'!
requestDequeueTime: anInteger
	requestDequeueTime := anInteger! !

!MaServerRequestLifecycle methodsFor: 'accessing' stamp: 'cmm 10/28/2008 14:48'!
requestObject
	^ requestObject! !

!MaServerRequestLifecycle methodsFor: 'accessing' stamp: 'cmm 10/28/2008 14:48'!
requestObject: aMaClientServerRequest
	requestObject := aMaClientServerRequest! !

!MaServerRequestLifecycle methodsFor: 'accessing' stamp: 'cmm 9/29/2008 22:31'!
requestQueueTime
	^ requestQueueTime! !

!MaServerRequestLifecycle methodsFor: 'initialize-release' stamp: 'cmm 5/1/2008 16:24'!
requestQueueTime: anInteger
	requestQueueTime := anInteger! !

!MaServerRequestLifecycle methodsFor: 'accessing' stamp: 'cmm 5/1/2008 19:30'!
responseByteArray
	^responseByteArray! !

!MaServerRequestLifecycle methodsFor: 'accessing' stamp: 'cmm 5/1/2008 19:30'!
responseByteArray: aByteArray 
	responseByteArray := aByteArray! !

!MaServerRequestLifecycle methodsFor: 'accessing' stamp: 'cmm 5/1/2008 16:18'!
socket
	^socket! !

!MaServerRequestLifecycle methodsFor: 'initialize-release' stamp: 'cmm 11/9/2008 20:13'!
socket: aSocket 
	socket := aSocket.
	socketQueueTime := Time millisecondClockValue.
	remoteAddress := socket remoteAddress.
	remotePort := socket remotePort! !

!MaServerRequestLifecycle methodsFor: 'accessing' stamp: 'cmm 5/1/2008 16:41'!
timeToExtractRequestBytes
	^timeToExtractRequestBytes! !

!MaServerRequestLifecycle methodsFor: 'accessing' stamp: 'cmm 5/1/2008 16:42'!
timeToExtractRequestBytes: anInteger
	"Specified in anInteger milliseconds."
	timeToExtractRequestBytes := anInteger! !

!MaServerRequestLifecycle methodsFor: 'accessing' stamp: 'cmm 5/1/2008 19:27'!
timeToSendResponse
	^ Time 
		milliseconds: doneSendingResponseTime
		since: beginSendingResponseTime! !

!MaServerRequestLifecycle methodsFor: 'accessing' stamp: 'cmm 5/1/2008 17:58'!
timeWaitingInRequestQueue
	^ Time 
		milliseconds: requestDequeueTime
		since: requestQueueTime! !

!MaServerRequestLifecycle methodsFor: 'accessing' stamp: 'cmm 5/1/2008 17:42'!
timeWaitingInSocketQueue
	^ Time 
		milliseconds: socketDequeueTime
		since: socketQueueTime! !

MaObject subclass: #MaServerSocket
	instanceVariableNames: 'socketQueue requestQueue processRequestProcess socket timeoutSeconds doBlock port running server socketQueueMonitorProcess clientConnections clientConnectionsGuard preferences networkProtocol interface'
	classVariableNames: 'ActiveSockets'
	poolDictionaries: ''
	category: 'Ma client server-Server API'!

!MaServerSocket commentStamp: 'cmm 6/6/2008 10:19' prior: 0!
I am the class which manages the underlying Squeak socket that listens for connecting client sockets.  I receive and process requests in the order received in the background and send responses off in their own background process as well.!

!MaServerSocket class methodsFor: 'initializing' stamp: 'cmm 11/21/2008 12:52'!
initialize
	ActiveSockets ifNotNil: [ ActiveSockets do: [ : each | each shutdown ] ].
	ActiveSockets := Set new.
	Smalltalk
		addToStartUpList: self ;
		addToShutDownList: self! !

!MaServerSocket class methodsFor: 'initializing' stamp: 'cmm 12/27/2004 23:16'!
shutDown: aboutToQuit

	super shutDown: aboutToQuit.
	ActiveSockets do: [ :each | each primShutdown ]! !

!MaServerSocket class methodsFor: 'initializing' stamp: 'cmm 2/10/2005 09:42'!
startUp: amResuming

	super startUp: amResuming.
	ActiveSockets do: 
		[ :each | 
		"If another startUp method creates and starts one of me listening, then I don't want to tell it to listen again"
		each isRunning ifFalse: [ each listen ] ]! !

!MaServerSocket methodsFor: 'private' stamp: 'cmm 7/5/2008 23:01'!
byteArrayToSendFor: aByteArray 
	| answer byteArrayToSend sendSize header |
	byteArrayToSend := aByteArray size > self compressThreshold  
		ifTrue: 
			[ | compressed |
			compressed := aByteArray lzCompressed.
			compressed size < aByteArray size 
				ifTrue: [ compressed ]
				ifFalse: [ aByteArray ] ]
		ifFalse: [ aByteArray ].
	answer := ByteArray new: byteArrayToSend size + self headerSize.
	sendSize := answer size.
	"Set high bit if compressed."
	header := self 
		setCompressedBit: aByteArray ~~ byteArrayToSend
		for: answer size.
	answer 
		maUint: 32
		at: 0
		put: header.
	answer 
		replaceFrom: self headerSize + 1
		to: sendSize
		with: byteArrayToSend
		startingAt: 1.
	^ answer! !

!MaServerSocket methodsFor: 'accessing' stamp: 'cmm 7/5/2008 22:53'!
compressThreshold
	"The maximum number of bytes a request will be before the client decides it should be compressed."
	^ preferences compressThreshold ! !

!MaServerSocket methodsFor: 'accessing' stamp: 'cmm 7/5/2008 22:53'!
compressThreshold: anInteger 
	"The maximum number of bytes a request will be before the client decides it should be compressed."
	preferences compressThreshold: anInteger! !

!MaServerSocket methodsFor: 'private' stamp: 'cmm 5/1/2008 18:05'!
dequeueByteArrayRequest
	| requestLifecycle |
	requestLifecycle := requestQueue next.
	requestLifecycle requestDequeueTime: Time millisecondClockValue.
	server noteRequestDequeued: requestLifecycle.
	^ requestLifecycle! !

!MaServerSocket methodsFor: 'private' stamp: 'cmm 11/9/2008 19:42'!
extractByteArrayRequestsFrom: aMaServerRequestLifecycle 
	"A socket connection has just been established.  Kickoff a new iterating process for the purpose of listening for requests bytes from a single client.  receiveOneByteArrayRequestFrom: waits a few seconds on the port, if data comes in it will extract the entire request and return it.  Otherwise it returns nil and the loop can check for shutdown or expiry."
	clientConnectionsGuard critical: 
		[ 1 
			to: clientConnections size - self maxConnections
			do: [ : n | (clientConnections at: n) shouldContinue: false ].
		clientConnections add: (MaClientConnection 
				serverSocket: self
				lifecycle: aMaServerRequestLifecycle) monitorForByteArrayRequests.
		server noteNumberOfConnections: clientConnections size ]! !

!MaServerSocket methodsFor: 'private' stamp: 'cmm 9/24/2002 23:01'!
headerSize
	"4-byte size header."

	^4! !

!MaServerSocket methodsFor: 'printing' stamp: 'cmm 7/16/2008 09:55'!
hostAndPortStringFor: aSocket 
	^ 'host: ' , (NetNameResolver stringFromAddress: aSocket remoteAddress) , ' port: ' , aSocket port printString! !

!MaServerSocket methodsFor: 'initialize-release' stamp: 'cmm 11/12/2008 10:33'!
initialize
	super initialize.
	running := false.
	self initializeQueues.
	timeoutSeconds := 30.
	self server: MaServer new.	"use the do-nothing server by default"
	doBlock := [ : lifecycle | server processRequestByteArray: lifecycle ].
	clientConnections := OrderedCollection new.
	clientConnectionsGuard := Mutex new.
	preferences := MaClientServerPreferences new.
	networkProtocol := #tcp.
	interface := NetNameResolver addressFromString: '0.0.0.0'! !

!MaServerSocket methodsFor: 'private' stamp: 'cmm 4/30/2003 00:30'!
initializeQueues

	socketQueue := SharedQueue new.
	requestQueue := SharedQueue new! !

!MaServerSocket methodsFor: 'accessing' stamp: 'cmm 11/11/2008 20:50'!
interface
	^interface! !

!MaServerSocket methodsFor: 'accessing' stamp: 'cmm 11/11/2008 20:51'!
interface: byteArrayAddress
	"Specify which network interface this socket should listen/read."
	interface := byteArrayAddress! !

!MaServerSocket methodsFor: 'testing' stamp: 'cmm 2/10/2005 09:42'!
isRunning

	^ running! !

!MaServerSocket methodsFor: 'testing' stamp: 'cmm 11/11/2008 14:15'!
isUdp
	^ networkProtocol == #udp! !

!MaServerSocket methodsFor: 'private' stamp: 'cmm 12/21/2008 18:43'!
listen
	"By using this method, I will call my servers #processRequestByteArray:.  If you don't want to have a MaServer class, use listenOn:answer:."
	running ifTrue: [ MaClientServerUserError signal: 'socket already listening.' ].
	running := true.
	ActiveSockets add: self.
	server note: 'creating server socket'.
	self isUdp 
		ifTrue: 
			[ socket := Socket newUDP
				setPort: port ;
				maPrimConnectTo: interface yourself ]
		ifFalse: 
			[ socket := Socket newTCP.
			socket 
				listenOn: port
				backlogSize: 50
				interface: interface ].
	socket isValid ifFalse: [ NetworkError signal: 'Socket accept() not supported.' ].
	server
		note: 'now listening on ' , port printString ;
		noteAboutToStart.
	self isUdp 
		ifTrue: 
			[ clientConnections add: (MaClientConnection 
					serverSocket: self
					lifecycle: (MaServerRequestLifecycle socket: socket)) monitorForUdpByteArrayRequests ]
		ifFalse: 
			[ self
				startAcceptingRequests ;
				startExtractingRequests ].
	self startProcessingRequests.
	server noteJustStarted! !

!MaServerSocket methodsFor: 'private' stamp: 'cmm 6/8/2008 16:51'!
maxConnections
	^ 50! !

!MaServerSocket methodsFor: 'accessing' stamp: 'cmm 11/11/2008 14:15'!
networkProtocol
	"Answer #tcp for TCP/IP interfaces, #udp for UDP."
	^ networkProtocol! !

!MaServerSocket methodsFor: 'accessing' stamp: 'cmm 11/11/2008 16:27'!
networkProtocol: aSymbol 
	"Specify #tcp or #udp."
	networkProtocol := aSymbol! !

!MaServerSocket methodsFor: 'accessing' stamp: 'cmm 9/21/2008 16:16'!
port
	^ socket ifNotNil: [ socket port ]! !

!MaServerSocket methodsFor: 'private' stamp: 'cmm 11/10/2008 12:04'!
primShutdown
	running := false.
	server note: 'new sockets no longer accepted.'.
	socketQueue isEmpty ifFalse: [ server note: 'waiting for queued sockets..' ].
	[ socketQueue isEmpty ] whileFalse: [ (Delay forSeconds: 1) wait ].
	socketQueueMonitorProcess ifNotNilDo: [ : proc | proc terminate ].
	requestQueue isEmpty ifFalse: [ server note: 'waiting for pending requests to process..' ].
	[ processRequestProcess isTerminated or: [ requestQueue isEmpty ] ] whileFalse: [ (Delay forSeconds: 1) wait ].
	processRequestProcess terminate.
	server note: 'deallocating resources.'.
	socket ifNotNil: [ socket closeAndDestroy: self timeoutSeconds ].
	self initializeQueues.
	socket := nil! !

!MaServerSocket methodsFor: 'start/stop' stamp: 'cmm 11/10/2008 16:03'!
processOn: portNumber 
	"By using this method, I will call my servers #processRequestByteArray:.  If you don't want to have a MaServer class, use listenOn:answer:."
	port := portNumber.
	self listen! !

!MaServerSocket methodsFor: 'start/stop' stamp: 'cmm 11/10/2008 16:05'!
processOn: portNumber using: oneArgBlock 
	"Listen on portNumber, for each ByteArray request received answer the value of oneArgBlock (which should answer a ByteArray)."
	doBlock := oneArgBlock.
	self processOn: portNumber! !

!MaServerSocket methodsFor: 'private' stamp: 'cmm 11/10/2008 09:52'!
processRequest: aMaServerRequestLifecycle 
	| responseByteArray |
	responseByteArray := doBlock value: aMaServerRequestLifecycle.
	server noteLastRequestFinish: (requestQueue isEmpty ifTrue: [ Time millisecondClockValue ]).
	(self isUdp not and: [ aMaServerRequestLifecycle requestObject wantsResponse ]) ifTrue: 
		[ self 
			sendData: responseByteArray
			using: aMaServerRequestLifecycle ]! !

!MaServerSocket methodsFor: 'private' stamp: 'cmm 12/23/2008 19:30'!
queueRequest: aMaRequestLifecycle 
	running ifTrue: 
		[ aMaRequestLifecycle requestQueueTime: Time millisecondClockValue.
		server noteRequestsQueued: requestQueue size.
		requestQueue nextPut: aMaRequestLifecycle.
		server noteRequestQueued: aMaRequestLifecycle ]! !

!MaServerSocket methodsFor: 'private' stamp: 'cmm 6/2/2008 20:19'!
queueSocket: aSocket 
	socketQueue nextPut: (MaServerRequestLifecycle socket: aSocket).
	server noteSocketQueued: aSocket! !

!MaServerSocket methodsFor: 'private' stamp: 'cmm 6/8/2008 11:34'!
removeConnection: aMaClientConnection 
	clientConnectionsGuard critical: 
		[ clientConnections 
			remove: aMaClientConnection
			ifAbsent: 
				[ "not worth blowing up"
				 ].
		server noteNumberOfConnections: clientConnections size ]! !

!MaServerSocket methodsFor: 'private' stamp: 'cmm 8/7/2008 00:22'!
sendData: aByteArray using: aMaRequestLifecycle 
	[ | byteArrayToSend destinationSocket |
	destinationSocket := aMaRequestLifecycle socket.
	destinationSocket isConnected 
		ifTrue: 
			[ server noteTimeToLoadBuffers: (Time millisecondsToRun: [ byteArrayToSend := self byteArrayToSendFor: aByteArray ]).
			aMaRequestLifecycle responseByteArray: byteArrayToSend.
			aMaRequestLifecycle beginSendingResponse.
			server 
				noteSendingResponse: byteArrayToSend
				to: destinationSocket.
			
			[ destinationSocket
				sendData: byteArrayToSend ;
				waitForSendDoneFor: timeoutSeconds ] 
				on: ConnectionTimedOut
				do: [ : exc | destinationSocket closeAndDestroy ].
			(destinationSocket isConnected and: [ destinationSocket sendDone ]) 
				ifTrue: 
					[ aMaRequestLifecycle doneSendingResponse.
					server noteResponseSent: aMaRequestLifecycle ]
				ifFalse: [ server noteClientTimedOut: aMaRequestLifecycle ] ]
		ifFalse: [ server noteClientTimedOut: aMaRequestLifecycle ] ] forkAt: Processor userInterruptPriority! !

!MaServerSocket methodsFor: 'private' stamp: 'cmm 6/8/2008 10:26'!
server
	^ server! !

!MaServerSocket methodsFor: 'server' stamp: 'cmm 12/30/2002 22:08'!
server: aMaServer
	"It is not required for the user to set this, a default one will be created."

	server _ aMaServer! !

!MaServerSocket methodsFor: 'private' stamp: 'cmm 6/29/2008 11:10'!
setCompressedBit: aBoolean for: headerInteger 
	^ aBoolean 
		ifTrue: 
			[ "Below is the correct implementation, but we would need a 2GB ByteArray for the simpler faster implementation below not to work.  2GB can't happen."
			"ba maUint: 32 at: 0 put: ((ba maUint: 32 at: 0) maBitSet: 32)."
			headerInteger maBitSet: 32 ]
		ifFalse: [ headerInteger maBitClear: 32 ]! !

!MaServerSocket methodsFor: 'start/stop' stamp: 'cmm 4/30/2003 00:46'!
shutdown

	ActiveSockets remove: self ifAbsent: [ "do nothing" ].
	self primShutdown! !

!MaServerSocket methodsFor: 'private' stamp: 'cmm 7/5/2008 22:54'!
startAcceptingRequests
	| clientSocket |
	
	[ [ running ] whileTrue: 
		[ clientSocket := socket 
			waitForAcceptFor: 5 "seconds to wait for server to shutdown, if asked"
			ifTimedOut: [ nil ].
		(clientSocket notNil and: [ clientSocket isConnected ]) ifTrue: [ self queueSocket: clientSocket ] ].
	server noteNoLongerAcceptingRequests ] newProcess
		name: 'Socket acceptor' ;
		priority: Processor userInterruptPriority ;
		resume.
	server note: 'Ready to queue sockets'! !

!MaServerSocket methodsFor: 'private' stamp: 'cmm 6/5/2008 23:21'!
startExtractingRequests
	socketQueueMonitorProcess := 
	[ [ running or: [ socketQueue notEmpty ] ] whileTrue: 
		[ | requestLifecycle |
		requestLifecycle := socketQueue next.
		requestLifecycle noteSocketDequeued.
		server noteSocketDequeued: requestLifecycle.
		self extractByteArrayRequestsFrom: requestLifecycle.
		requestLifecycle := nil ].
	server noteNoLongerExtractingRequests ] newProcess
		name: 'socketQueue Monitor' ;
		priority: Processor userInterruptPriority ;
		resume.
	server note: 'ready to extract requests'! !

!MaServerSocket methodsFor: 'private' stamp: 'cmm 12/22/2008 17:03'!
startProcessingRequests
	processRequestProcess := 
	[ [ running or: [ requestQueue notEmpty ] ] whileTrue: 
		[ [ self processRequest: self dequeueByteArrayRequest ] 
			on: Error
			do: 
				[ : err | 
				MaClientServerPreferences debug 
					ifTrue: [ err pass ]
					ifFalse: 
						[ "Ok, let's be extra aggressive in not letting server crash.  If one request generates an unexpected error, just log it and discard, but still process subsequent requests."
						Smalltalk 
							logError: err description
							inContext: err signalerContext
							to: 'SqueakDebug.log' ] ] ].
	server noteNoLongerProcessingRequests ] newProcess
		name: 'Request processor' ;
		resume.
	server note: 'ready to process requests'! !

!MaServerSocket methodsFor: 'accessing' stamp: 'cmm 7/5/2008 22:52'!
timeoutSeconds
	"The number of seconds to wait for responses from network requests before assuming they are lost."
	^ preferences timeoutSeconds! !

!MaServerSocket methodsFor: 'accessing' stamp: 'cmm 7/5/2008 22:52'!
timeoutSeconds: anInteger 
	"The number of seconds to wait for responses from network requests before assuming they are lost."
	preferences timeoutSeconds: anInteger! !

MaSerializationStat subclass: #MaClientStat
	instanceVariableNames: 'timeToConnectSocketHistory timeToLoadBuffersHistory timeToSendRequestsHistory timeWaitingForResponsesHistory timeReceivingResponsesHistory timeToSubmitByRequestClassHistorys'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma client server-Client API'!

!MaClientStat methodsFor: 'enumerate' stamp: 'cmm 5/9/2008 17:46'!
historysDo: aBlock 
	super historysDo: aBlock.
	{ 
		timeToConnectSocketHistory.
		timeToLoadBuffersHistory.
		timeToSendRequestsHistory.
		timeWaitingForResponsesHistory.
		timeReceivingResponsesHistory
	 } do: aBlock.
	timeToSubmitByRequestClassHistorys do: aBlock! !

!MaClientStat methodsFor: 'initialize-release' stamp: 'cmm 5/9/2008 17:45'!
initialize
	super initialize.
	timeToSubmitByRequestClassHistorys := Dictionary new.
	timeToConnectSocketHistory := MaStatHistory name: 'Time to Connect Sockets to Server'.
	timeToLoadBuffersHistory := MaStatHistory name: 'Time to Load Buffers'.
	timeToSendRequestsHistory := MaStatHistory name: 'Time to Send Requests'.
	timeWaitingForResponsesHistory := MaStatHistory name: 'Time for Initial Response'.
	timeReceivingResponsesHistory := MaStatHistory name: 'Time to Download Full Response'! !

!MaClientStat methodsFor: 'copying' stamp: 'cmm 12/22/2008 19:54'!
postArchivalCopy
	timeToSubmitByRequestClassHistorys := timeToSubmitByRequestClassHistorys copy.
	super postArchivalCopy! !

!MaClientStat methodsFor: 'historys' stamp: 'cmm 5/9/2008 17:32'!
timeReceivingResponsesHistory
	"The time from when the initial response is received from the server to when all bytes of the response are downloaded.  By excluding the time waiting for the *initial* response, this stat attempts to measure the speed of downloading responses over the network."
	^timeReceivingResponsesHistory! !

!MaClientStat methodsFor: 'historys' stamp: 'cmm 5/9/2008 17:33'!
timeToConnectSocketHistory
	"Time to initiate a socket-connections, performed for for each request of the server."
	^timeToConnectSocketHistory! !

!MaClientStat methodsFor: 'historys' stamp: 'cmm 5/9/2008 17:33'!
timeToLoadBuffersHistory
	"For now Ma client server makes lots of copies of ByteArrays.  Until this is optimized, this stat tracks it to see how big of an impact it is."
	^timeToLoadBuffersHistory! !

!MaClientStat methodsFor: 'historys' stamp: 'cmm 5/9/2008 17:34'!
timeToSendRequestsHistory
	"The time needed just to send requests to the server."
	^timeToSendRequestsHistory! !

!MaClientStat methodsFor: 'historys' stamp: 'cmm 6/25/2008 14:43'!
timeToSubmitHistoryFor: requestClass 
	"The total time, by request class, including serialization, materialization, and all other house keeping.  A useful set of stats measuring the total reality for the client, by request type."
	^ timeToSubmitByRequestClassHistorys 
		at: requestClass
		ifAbsentPut: 
			[ MaStatHistory new
				name: 'Time waiting for ' , requestClass name , ' requests' ;
				transition ;
				yourself ]! !

!MaClientStat methodsFor: 'historys' stamp: 'cmm 5/9/2008 17:40'!
timeWaitingForResponsesHistory
	"When all bytes for a request have been pumped into the TCP/IP socket connected to the server, this stat immediately starts timing until we get the *first* initial response bytes back (which won't occur until the server has completely processed our request and is now sending back the response).  Therefore, this stat is a approximate measure of the time the server spends processing each request, including any time waiting in its request queue."
	^timeWaitingForResponsesHistory! !

MaSerializationStat subclass: #MaServerStat
	instanceVariableNames: 'startTime stopTime requestsQueuedHistory socketsQueuedHistory requestSizeHistory requestProcessingTimeHistory requestProcessingTimeHistorys responseSizeHistory extractTimeHistory timeWaitingInSocketQueueHistory timeWaitingInRequestQueueHistory timeToSendResponsesHistory clientTimeoutsHistory numberOfClientConnectionsHistory timeToLoadBuffersHistory idleTimeHistory'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma client server-Server API'!

!MaServerStat methodsFor: 'historys' stamp: 'cmm 5/7/2008 17:28'!
clientTimeoutsHistory
	"This history describes how often clients timed out.  When a request completes its lifecycle with no errors, a score of 0 is given.  When a network error occurs receiving the request or sending the response, a score of 1 is added.  Therefore, the sum is the total number of timeouts, #mean is how often they timed out."
	^ clientTimeoutsHistory! !

!MaServerStat methodsFor: 'recording' stamp: 'cmm 5/7/2008 17:29'!
decrementSocketsQueued
	| stat |
	stat := self socketsQueuedStat.
	stat last: stat last - 1! !

!MaServerStat methodsFor: 'historys' stamp: 'cmm 5/7/2008 17:29'!
extractTimeHistory
	"This history describes the amount of time, in milliseconds, elapsed from the initial client socket connection to the point in which the entire request is uploaded to my server."
	^ extractTimeHistory! !

!MaServerStat methodsFor: 'private' stamp: 'cmm 9/29/2008 21:43'!
historysDo: aBlock 
	super historysDo: aBlock.
	{ 
		socketsQueuedHistory.
		timeWaitingInSocketQueueHistory.
		extractTimeHistory.
		requestsQueuedHistory.
		timeWaitingInRequestQueueHistory.
		requestSizeHistory.
		requestProcessingTimeHistory.
		responseSizeHistory.
		timeToSendResponsesHistory.
		clientTimeoutsHistory.
		numberOfClientConnectionsHistory.
		timeToLoadBuffersHistory.
		idleTimeHistory
	 } do: aBlock.
	requestProcessingTimeHistorys do: aBlock! !

!MaServerStat methodsFor: 'historys' stamp: 'cmm 9/29/2008 22:01'!
idleTimeHistory
	"The time the server waited after processing a request, before a new request came in."
	^ idleTimeHistory! !

!MaServerStat methodsFor: 'recording' stamp: 'cmm 5/7/2008 17:45'!
incrementSocketsQueued
	| history |
	history := self socketsQueuedHistory.
	history addValue: history last + 1! !

!MaServerStat methodsFor: 'initialize-release' stamp: 'cmm 12/23/2008 18:55'!
initialize
	super initialize.
	socketsQueuedHistory := MaStatHistory name: 'Sockets Queued'.
	requestsQueuedHistory := MaStatHistory name: 'Requests Queued'.
	requestSizeHistory := MaStatHistory name: 'Request Sizes'.
	requestProcessingTimeHistory := MaStatHistory name: 'Request Processing Times'.
	requestProcessingTimeHistorys := Dictionary new.
	responseSizeHistory := MaStatHistory name: 'Response Sizes'.
	extractTimeHistory := MaStatHistory name: 'Request Upload Times'.
	timeWaitingInSocketQueueHistory := MaStatHistory name: 'Time Waiting in Socket Queue'.
	timeWaitingInRequestQueueHistory := MaStatHistory name: 'Time Waiting in Request Queue'.
	timeToSendResponsesHistory := MaStatHistory name: 'Response Upload Times'.
	clientTimeoutsHistory := MaStatHistory name: 'Failed Lifecycles'.
	numberOfClientConnectionsHistory := MaStatHistory name: 'Number of Connected Client Sockets'.
	timeToLoadBuffersHistory := MaStatHistory name: 'Time to Compress and Uncompress'.
	idleTimeHistory := MaStatHistory name: 'Server Idle Times'.
	startTime := DateAndTime now! !

!MaServerStat methodsFor: 'initialize-release' stamp: 'cmm 12/22/2008 19:54'!
noteStop
	stopTime := DateAndTime now! !

!MaServerStat methodsFor: 'historys' stamp: 'cmm 6/8/2008 11:33'!
numberOfClientConnectionsHistory
	"The number of client sockets simultaneously connected to this server.  There is a Process running for each one that extracts the request bytes from them."
	^numberOfClientConnectionsHistory! !

!MaServerStat methodsFor: 'copying' stamp: 'cmm 12/22/2008 19:56'!
postArchivalCopy
	requestProcessingTimeHistorys := requestProcessingTimeHistorys copy.
	super postArchivalCopy.
	self noteStop! !

!MaServerStat methodsFor: 'recording' stamp: 'cmm 6/5/2008 21:58'!
recordClientTimeout
	self clientTimeoutsHistory addValue: 1! !

!MaServerStat methodsFor: 'recording' stamp: 'cmm 9/29/2008 15:28'!
recordIdleTime: milliseconds
	self idleTimeHistory addValue: milliseconds! !

!MaServerStat methodsFor: 'recording' stamp: 'cmm 6/5/2008 21:58'!
recordRequestSize: anInteger 
	self requestSizeHistory addValue: anInteger! !

!MaServerStat methodsFor: 'recording' stamp: 'cmm 6/5/2008 21:59'!
recordResponseSize: anInteger 
	self responseSizeHistory addValue: anInteger! !

!MaServerStat methodsFor: 'recording' stamp: 'cmm 6/5/2008 21:59'!
recordSuccessfulLifecycle
	self clientTimeoutsHistory addValue: 0! !

!MaServerStat methodsFor: 'recording' stamp: 'cmm 6/5/2008 21:59'!
recordTimeToExtractRequestBytes: milliseconds 
	self extractTimeHistory addValue: milliseconds! !

!MaServerStat methodsFor: 'recording' stamp: 'cmm 6/5/2008 21:59'!
recordTimeToSendResponse: anInteger 
	self timeToSendResponsesHistory addValue: anInteger! !

!MaServerStat methodsFor: 'recording' stamp: 'cmm 6/5/2008 21:59'!
recordTimeWaitingInRequestQueue: anInteger
	self timeWaitingInRequestQueueHistory addValue: anInteger! !

!MaServerStat methodsFor: 'recording' stamp: 'cmm 6/5/2008 21:59'!
recordTimeWaitingInSocketQueue: anInteger
	self timeWaitingInSocketQueueHistory addValue: anInteger! !

!MaServerStat methodsFor: 'historys' stamp: 'cmm 5/7/2008 17:33'!
requestProcessingTimeHistory
	"These statistics represent the time, in milliseconds, actually consumed by the Magma server to process client requests.  Although requests are processed one at a time, responses are sent in separate threads and request input downloads are also concurrent."
	^ requestProcessingTimeHistory! !

!MaServerStat methodsFor: 'historys' stamp: 'cmm 6/25/2008 11:03'!
requestProcessingTimeHistoryFor: requestObject 
	"Different requests take different amounts of time to process.  Track them by the class of request."
	^ requestProcessingTimeHistorys 
		at: requestObject class
		ifAbsentPut: 
			[ MaStatHistory new
				name: 'Time to process ' , requestObject class name ;
				transition ;
				yourself ]! !

!MaServerStat methodsFor: 'historys' stamp: 'cmm 5/7/2008 17:33'!
requestSizeHistory
	"The size of client requests, in bytes.  This would be related to extractTime, timeToMaterializeRequests and requestProcessingTimes."
	^ requestSizeHistory! !

!MaServerStat methodsFor: 'historys' stamp: 'cmm 5/7/2008 17:34'!
requestsQueuedHistory
	"The number of requests queued at each moment a new client request is queued.  How long does the request line get.."
	^ requestsQueuedHistory! !

!MaServerStat methodsFor: 'accessing' stamp: 'cmm 5/7/2008 17:32'!
requestsQueuedStat
	^ requestsQueuedHistory active! !

!MaServerStat methodsFor: 'historys' stamp: 'cmm 5/7/2008 17:34'!
responseSizeHistory
	"A statistic about how large the responses are.  As these numbers increase so, too, should the timeToSerializeResponses and timeToSendResponses increase."
	^ responseSizeHistory! !

!MaServerStat methodsFor: 'historys' stamp: 'cmm 5/7/2008 17:34'!
socketsQueuedHistory
	"The number of sockets in a queue waiting solely to have the bytes of the request extracted from the socket."
	^ socketsQueuedHistory! !

!MaServerStat methodsFor: 'accessing' stamp: 'cmm 5/7/2008 17:32'!
socketsQueuedStat
	^ socketsQueuedHistory active! !

!MaServerStat methodsFor: 'accessing' stamp: 'cmm 5/7/2008 17:32'!
startTime

	^ startTime! !

!MaServerStat methodsFor: 'accessing' stamp: 'cmm 5/7/2008 17:32'!
startTime: aMaCalendarTime

	startTime _ aMaCalendarTime! !

!MaServerStat methodsFor: 'accessing' stamp: 'cmm 5/7/2008 17:32'!
stopTime

	^ stopTime! !

!MaServerStat methodsFor: 'accessing' stamp: 'cmm 5/7/2008 17:32'!
stopTime: aMaCalendarTime

	stopTime _ aMaCalendarTime! !

!MaServerStat methodsFor: 'historys' stamp: 'cmm 6/8/2008 21:33'!
timeToLoadBuffersHistory
	^timeToLoadBuffersHistory! !

!MaServerStat methodsFor: 'historys' stamp: 'cmm 5/7/2008 17:34'!
timeToSendResponsesHistory
	"The time required to send serialized responses out the TCP/IP port over the network.  Responses are sent in individual background threads."
	^ timeToSendResponsesHistory! !

!MaServerStat methodsFor: 'historys' stamp: 'cmm 5/7/2008 17:34'!
timeWaitingInRequestQueueHistory
	"The time, in milliseconds, from the point when the request is put into the request queue, all requests in front of it have been processed, and it is dequeued."
	^ timeWaitingInRequestQueueHistory! !

!MaServerStat methodsFor: 'historys' stamp: 'cmm 5/7/2008 17:34'!
timeWaitingInSocketQueueHistory
	"The duration, in milliseconds, from when the socket is queued until extraction of its ByteArray can begin.  All sockets in front of it get their ByteArray's extracted first."
	^ timeWaitingInSocketQueueHistory! !

!Object methodsFor: '*ma client server' stamp: 'cmm 10/28/2008 16:36'!
wantsResponse
	"Until protocol is established, we must only use types of objects that are part of the original serializer.  The initial request for protocol establishment is a simple hard-coded String."
	^ true! !

MaError subclass: #MaMalformedRequestError
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma client server-Exceptions'!

!MaMalformedRequestError commentStamp: 'cmm 2/1/2005 15:26' prior: 0!
I am signaled when a request is received from a client that could not be materialized.!

PackageInfo subclass: #MaClientServerPackageInfo
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma client server-Info'!

!MaClientServerPackageInfo class methodsFor: 'as yet unclassified' stamp: 'cmm 12/27/2008 13:27'!
packageName
	^ 'Ma client server'! !

!MaClientServerPackageInfo methodsFor: 'as yet unclassified' stamp: 'cmm 1/31/2003 00:20'!
additionalFiles

	^#('clientTest.ws' 'serverTest.ws')! !

!MaClientServerPackageInfo methodsFor: 'as yet unclassified' stamp: 'cmm 4/8/2004 23:13'!
changesText

	^ '
version 5:
	updated for 3.6 and 3.7.  This version does not work with 3.5.

version 4:
	running servers now survive image restarts.  The network is initialized by MaClientSocket upon starting the image, allowing sockets to immediately reconnect upon image start up.

version 3:
  performance improvement
  support for multiple clients in the same image connected to the same remote server

version 2:
  Should now support big and little endian machines simultaneously.

version 1:
  Initial public release.'! !

!MaClientServerPackageInfo methodsFor: 'accessing' stamp: 'cmm 12/27/2008 13:49'!
prerequisitePackageNames
	^ #('Ma object serialization' )! !

!MaClientServerPackageInfo methodsFor: 'as yet unclassified' stamp: 'cmm 1/30/2003 22:54'!
readmeText

	^'!!!!Ma client server

!!!!Introduction
I can help you write a client-server program very easily.

In the client-server model, a client sends a request to a server, the server processes the request and sends back a response.

This framework provides two layers interoperation.  You choose which layer you want to use and then write two programs.  A client program that will send requests to a remote server, which is running the other program to process those requests.

If you choose the lower layer, the "socket layer," you send ByteArray requests and receive ByteArray responses.  In the upper layer, the "object request layer" you send arbitrary request objects and get arbitrary objects as responses.  The "Ma object serialization" framework is used to convert the objects to ByteArrays which are sent and received for you using the socket layer.

!!!!Using the Socket Layer

!!Setting up the server
If you just need to send and receive ByteArray''s, you write your server program any way you want.  Just keep in mind that, somewhere, you''re going to want to take-in a ByteArray as input and you must answer a ByteArray as the response.

Then, to start the server:

=	myServerSocket _ MaServerSocket new
=		listenOn: 12345
=		answer:
=			[ :requestByteArray |
=			... give requestByteArray to your server program ... ]

MaServerSocket utilizes SharedQueue''s and background Processes to ensure that request processing is not held up by sending or receiving the ByteArray''s, which occurs entirely in the background.  The answer block is only ever evaluated for one client request at a time.

When you''re done, be sure to shutdown the socket so resources are properly deallocated:

=	myServerSocket shutdown

!!Getting answers in the client
Your client program creates a MaClientSocket by specifying the ip address as numbers in a ByteArray:

=	mySocket _ MaClientSocket hostAddress: #(192 168 1 1) port: 12345.

No connections are made just from creating the socket.  That happens when you actually send the request:

=	responseByteArray _ mySocket sendData: myRequestByteArray

There''s also a sendData:startingAt:count:waitForReplyIn: method for maximum efficiency.  If the reply ByteArray you supply wasn''t large enough, you''ll get a bigger one.

!!Socket Events
MaServerSocket knows about a "server", which defaults to a Server that does nothing (MaServer).  But you subclass MaServer and override some of its methods to learn when a socket is queued, a ByteArray request has been fully delivered, or sending a response, etc.  See MaServer for more information.

!!!!Using the Object Request Layer
If you choose to use the object request layer, you send "request" objects and get back arbitrary response objects.  So you write your client and server programs the same except you don''t concern yourself with parsing ByteArrays anymore.

!!Setting up the server
When you are ready to start your server program, instead of creating a MaServerSocket, you create a MaTcpRequestServer:

=	myServer _ MaTcpRequestServer protocol: { MyRequestClass1. MyRequestClass2.  MyResponseClass1.  MyResponseClass2.  OrderedCollection.  Dictionary. "etc." }.

What''s this?  Yes, you have to tell it all the classes that you''ll be sending and receiving, accounting for the full object-graph of your requests.  For example, if you have SubmitArticleRequest that has an Article object, you must include not only SubmitArticleRequest, Article and all the other kinds of objects that Article references, and what they reference, and so on.  You don''t however, need to ever specify String, ByteArray, Set, Symbol or Array.  Those are automatically part of every protocol.

You may also optionally specify a MaServerConsole to receive events.  Subclass MaServerConsole and override one of its notification methods and it''ll be called if you:

=	myServer _ console: myConsole

When you''re ready to start the server, you use processOn:do:

=	myServer
=		processOn: 12345
=		do:
=			[ :requestObject |
=			... pass requestObject to your server program, which answers the response ... ]

When you''re done, you need to shutdown to release operating system resources:

=	myServer shutdown

!!Setting up the client
Your client program creates a MaTcpRequestServerLink pointing to the server hostAddress and port is is listening:

=	myLinkToServer _ MaTcpRequestServerLink
=		hostAddress: #(192 168 1 1)
=		port: 12345

Before you can do anything, you have to connect.  This will ensure you have all the needed protocol in this client image (and throw an exception if you don''t).

=	myLinkToServer connect

To balance network fluctuations against your patience, you may wish to override the default timeout period from 30 seconds:

=	myLinkToServer timeoutSeconds: 30

!!Getting answers in the client
Now that you''re connected, your client program can get answers from the server by:

=	myLinkToServer submit: someRequestObject

Remember, someRequestObject will be a member of the protocol, as will the response you get back from calling submit:.

That''s it!!  When you''re done, don''t forget to disconnect so as you release network resources back to the operating system.

=	myLinkToServer disconnect
'! !

MaUserError subclass: #MaClientServerUserError
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma client server-Exceptions'!

!Socket methodsFor: '*ma client server' stamp: 'cmm 11/11/2008 21:18'!
maPrimConnectTo: addressByteArray 
	self 
		primSocket: socketHandle
		connectTo: addressByteArray
		port: self port! !
MaClientServerPreferences initialize!
MaClientSocket initialize!
MaServerSocket initialize!
